# ---------------------------------------------------------------------
# Copyright (c) 2024 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------

from __future__ import annotations

import configparser
import os
from typing import Optional

import qai_hub as hub
from qai_hub.client import Client as HubClient
from qai_hub.public_rest_api import ClientConfig as HubClientConfig
from qai_hub.public_rest_api import _str2bool as str2bool

"""
A mapping of Hub clients for the given deployment names.
The mapping is Map<AI Hub deployment name (lower case), AI Hub client>

Deployment names are mapped to None in the dictionary if a client for them could not be created.
"""
_CACHED_CLIENTS: dict[str, HubClient | None] = {}


def _get_global_client() -> Optional[tuple[str, HubClient]]:
    """
    Returns [global client Hub deployment name], [global client]
    """
    try:
        global_client = hub.hub._global_client
        # The deployment name is the subdomain of AIHub that is used by the config.
        # Transform https://blah.aihub.qualcomm.com/ -> blah
        deployment_name = global_client.config.api_url.split("/")[2].split(".")[0]

        #
        # Prod deployment is a special case where the deployment "name" does not match the URL
        # "prod" deployment is "app.aihub.qualcomm.com"
        #
        deployment_name = (
            "prod" if deployment_name == "app" else deployment_name.lower()
        )

        return deployment_name, global_client
    except FileNotFoundError:
        # no global config exists
        pass
    except IndexError:
        # Can't read the global config's api_url. Ignore it
        pass

    return None


if x := _get_global_client():
    _CACHED_CLIENTS.__setitem__(*x)


def _read_hub_config(path: str) -> Optional[HubClient]:
    if not os.path.exists(path):
        return None

    config = configparser.ConfigParser()
    config.read(path)
    try:
        client_config = config["api"]
        api_config = HubClientConfig(
            api_url=client_config["api_url"],
            web_url=client_config["web_url"],
            api_token=client_config["api_token"],
            verbose=(
                str2bool(client_config["verbose"])
                if "verbose" in client_config
                else True
            ),
        )
        return HubClient(api_config)
    except KeyError:
        pass

    return None


def get_hub_client(deployment_name: str = "prod") -> Optional[HubClient]:
    deployment_name = deployment_name.lower()
    deployment_name = "prod" if deployment_name == "app" else deployment_name

    # Return Cached client if applicable
    if deployment_name in _CACHED_CLIENTS:
        return _CACHED_CLIENTS.get(deployment_name, None)

    # Create client if their tokens are in the global environment.
    upper_deployment_name = deployment_name.upper()
    client = None
    if user_token := os.environ.get(f"HUB_USER_TOKEN_{upper_deployment_name}", None):
        #
        # Prod deployment is a special case where the deployment "name" does not match the URL
        # "prod" deployment is "app.aihub.qualcomm.com"
        #
        deployment_name_url = "app" if deployment_name == "prod" else deployment_name
        #

        api_url = os.environ.get(
            f"HUB_API_URL_{upper_deployment_name}",
            f"https://{deployment_name_url}.aihub.qualcomm.com/",
        )
        web_url = os.environ.get(f"HUB_WEB_URL_{upper_deployment_name}", api_url)

        client = HubClient(HubClientConfig(api_url, web_url, user_token, True))
    else:
        config_path = os.path.expanduser(f"~/.qai_hub/{deployment_name}.ini")
        client = _read_hub_config(config_path)

    _CACHED_CLIENTS[deployment_name] = client
    return client


def get_hub_client_or_raise(deployment_name: str = "prod") -> HubClient:
    hub_client = get_hub_client(deployment_name)
    if not hub_client:
        raise ValueError(
            f"Unable to create a client for Hub deployment {deployment_name}"
        )
    return hub_client
