# ---------------------------------------------------------------------
# Copyright (c) 2024 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------
from __future__ import annotations

import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

import numpy as np
from qai_hub.client import Client, CompileJob, InputSpecs, SourceModelType, UserError

from qai_hub_models.configs.info_yaml import MODEL_DOMAIN, MODEL_TAG, MODEL_USE_CASE
from qai_hub_models.scorecard.device import ScorecardDevice
from qai_hub_models.scorecard.internal.hub_clients import get_hub_client
from qai_hub_models.utils.base_config import BaseQAIHMConfig

DEFAULT_MODELS_DIR = Path(os.path.dirname(__file__)) / "models"


@dataclass
class ScorecardModelConfig(BaseQAIHMConfig):
    @staticmethod
    def source_model_type_from_yaml_string(string: str) -> SourceModelType:
        return SourceModelType[string.upper().replace(" ", "_")]

    @staticmethod
    def source_model_type_to_yaml_string(type: SourceModelType) -> str:
        return type.name.upper().replace("_", " ")

    @dataclass
    class ModelInput(BaseQAIHMConfig):
        # Input name
        name: str

        # Input shape
        shape: list[int]

        # Input data type
        dtype: np.dtype = np.dtype(np.float32)

        def to_dict(self, include_defaults=True, yaml_compatible=False):
            out = super()._complete_partial_dict(
                {"dtype": self.dtype.name} if yaml_compatible else dict(),
                include_defaults,
                yaml_compatible,
            )

            return out

        @classmethod
        def from_dict(cls, val_dict) -> ScorecardModelConfig.ModelInput:
            val_dict["dtype"] = np.dtype(getattr(np, val_dict["dtype"]))
            return super().from_dict(val_dict)

    # Model ID (same as name of config file)
    id: str

    # Model to execute, uploaded to AI Hub.
    # Must be shared with qaihm_bot@qti.qualcomm.com
    hub_model_id: str

    # Example Model Input Data, uploaded as a dataset to AI Hub.
    # Must be shared with qaihm_bot@qti.qualcomm.com
    hub_input_dataset_id: str

    # Type of Source Model
    type: SourceModelType

    # Qualcomm Sponsors (users or NTK groups)
    sponsors: list[str]

    # Why was this model added to the AI Hub Models scorecard? What is its relevance to your BU / Qualcomm?
    context: str

    # Where did this model come from? How was it and the sample data generated / exported?
    # This should have links for model code, data, and weights unless it was developed fully internally.
    source: str

    # The domain the model is used in such as computer vision, audio, etc.
    domain: MODEL_DOMAIN

    # What task the model is used to solve, such as object detection, classification, etc.
    use_case: MODEL_USE_CASE

    # Applicable model tags
    tags: list[MODEL_TAG]

    # Valid form factors for this model. Only these form factors are tested.
    form_factors: list[ScorecardDevice.FormFactor]

    # If set, this model is skipped entirely. This should be set to a string reason if the model is to be disabled.
    disabled_reason: Optional[str] = None

    # If False, skips this model on chipsets that support only quantized compute.
    fully_quantized: bool = False

    # Model input spec (applicable only for TorchScript models)
    input_specs: Optional[list[ScorecardModelConfig.ModelInput]] = None

    # Model output names (applicable only for TorchScript models)
    output_names: Optional[list[str]] = None

    # Inputs that are in channel-first (NCHW) format.
    # In other words, these inputs have a channel dimension that immediately follows the batch dimension.
    #
    # When compiling for some runtimes, to optimize performance,
    # the shapes of these inputs are changed to channel-last (NHWC).
    # In channel-last format, the channel dimension is the rightmost (last) dimension.
    channel_first_inputs: list[str] = field(default_factory=(list))

    # Outputs that are in channel-first (NCHW) format.
    # In other words, these outputs have a channel dimension that immediately follows the batch dimension.
    #
    # When compiling for some runtimes, to optimize performance,
    # the shapes of these outputs are changed to channel-last (NHWC).
    # In channel-last format, the channel dimension is the rightmost (last) dimension.
    channel_first_outputs: list[str] = field(default_factory=(list))

    # Model IDs by AI Hub deployment-- Map<Hub deployment name : Model ID>.
    # This field is managed by Continuous Integration and should not be modified by users.
    hub_model_ids_automated: dict[str, str] = field(default_factory=dict)

    # Example Inputs by AI Hub deployment-- Map<Hub deployment name : Dataset ID>.
    # This field is managed by Continuous Integration and should not be modified by users.
    hub_input_dataset_ids_automated: dict[str, str] = field(default_factory=dict)

    # Example Inputs by AI Hub deployment-- Map<Hub deployment name : Dataset ID>.
    # channel_first -> channel last transform is applied on inputs found in self.channel_first_inputs.
    #
    # This field is managed by Continuous Integration and should not be modified by users.
    hub_input_channel_last_dataset_ids_automated: dict[str, str] = field(
        default_factory=dict
    )

    def get_hub_api_input_specs(self) -> InputSpecs | None:
        return (
            {x.name: (tuple(x.shape), x.dtype.name) for x in self.input_specs}
            if self.input_specs
            else None
        )

    def to_dict(self, include_defaults=True, yaml_compatible=False):
        return super()._complete_partial_dict(
            {"type": ScorecardModelConfig.source_model_type_to_yaml_string(self.type)}
            if yaml_compatible
            else dict(),
            include_defaults,
            yaml_compatible,
        )

    @classmethod
    def from_dict(cls, val_dict) -> ScorecardModelConfig:
        val_dict["type"] = ScorecardModelConfig.source_model_type_from_yaml_string(
            val_dict["type"]
        )
        val_dict["domain"] = MODEL_DOMAIN.from_string(val_dict["domain"])
        val_dict["use_case"] = MODEL_USE_CASE.from_string(val_dict["use_case"])
        val_dict["tags"] = [MODEL_TAG.from_string(tag) for tag in val_dict["tags"]]
        if "all" in [x.lower() for x in val_dict["form_factors"]]:
            val_dict["form_factors"] = [x for x in ScorecardDevice.FormFactor]
        else:
            val_dict["form_factors"] = [
                ScorecardDevice.FormFactor.from_string(tag)
                for tag in val_dict["form_factors"]
            ]

        if input_specs := val_dict.get("input_specs", None):
            val_dict["input_specs"] = [
                ScorecardModelConfig.ModelInput.from_dict(x) for x in input_specs
            ]

        return super().from_dict(val_dict)

    @classmethod
    def from_scorecard_model_id(
        cls, model_id: str, models_dir: Path = DEFAULT_MODELS_DIR
    ):
        path = models_dir / f"{model_id}.yaml"
        if not path.exists():
            raise ValueError(f"Could not find config for model {model_id}")
        return cls.from_yaml(path)

    def to_scorecard_yaml(self, models_dir: Path = DEFAULT_MODELS_DIR):
        return self.to_yaml(models_dir / f"{self.id}.yaml")

    def validate(self) -> Optional[str]:
        if self.type in [SourceModelType.ONNX, SourceModelType.AIMET_ONNX]:
            for attr in ["input_specs", "output_names"]:
                if getattr(self, attr) is not None:
                    return f"{attr} should not be defined if source model type is ONNX."

        elif self.type == SourceModelType.TORCHSCRIPT:
            for attr in ["input_specs", "output_names"]:
                if getattr(self, attr) is None:
                    return (
                        f"{attr} must be defined if source model type is Torchscript."
                    )

        else:
            return f"Unsupported source model type: {self.type.name}"

        if (
            self.hub_input_dataset_ids_automated.keys()
            != self.hub_model_ids_automated.keys()
        ):
            return "Automated dataset IDs and model IDs must exist for the same set of Hub deployments."

        # Collect the list of hub deployments we need access to for testing
        hub_client_names = {"prod"}.union(self.hub_input_dataset_ids_automated.keys())

        # For each deployment, if a client configuration exists in this process' environment, get a config for it.
        # Some deployment credentials may not be available in this environment. If credentials aren't available,
        # we don't run validation on any assets related to that hub deployment.
        hub_clients = {x: get_hub_client(x) for x in hub_client_names}

        def _verify_model(model_id: str, deployment: str, client: Client) -> str | None:
            """
            Validate the model aligns with this config.

            This is a "do the best we can given the information we have" validator. To fully validate a model against this config,
            one would need to download & process each model to extract I/O shapes. Doing so is too expensive for this function.
            """
            try:
                model = client.get_model(model_id)
            except UserError:
                return f"Cannot access model ID {model_id} on AI Hub deployment {deployment}"

            if model.model_type != self.type:
                return f"Model type {self.type} is not the same as type as model {model_id} ({model.model_type}) in AI Hub deployment {deployment}"

            # If the model was compiled with AI Hub, we have some additional I/O shape info accessible to validate with
            if producer_job := model.producer:
                assert isinstance(producer_job, CompileJob)
                for input_name in self.channel_first_inputs:
                    if input_name not in producer_job.target_shapes.keys():
                        return f"Input {input_name} is defined in channel_first_inputs, but does not exist in the ONNX model."

                if self.fully_quantized:
                    for input_name, input_value in producer_job.target_shapes.items():
                        assert isinstance(input_value, tuple)
                        if input_value[1] in ["int32", "float32"]:
                            return f"Model is marked fully quantized, but input {input_name} has type {input_value[1]}"
            return None

        def _verify_dataset(
            dataset_id: str, deployment: str, client: Client
        ) -> str | None:
            """
            Validate the dataset aligns with this config.

            This is a "do the best we can given the information we have" validator. To fully validate a dataset against this config,
            one would need to download & process each dataset to extract I/O shapes. Doing so is too expensive for this function.
            """
            try:
                client.get_dataset(dataset_id)
            except UserError:
                return f"Cannot access dataset ID {dataset_id} on AI Hub deployment {deployment}"
            return None

        # If we have access to prod, verify we have access to the prod model & dataset IDs in this config.
        if client := hub_clients.get("prod", None):
            if bad_val := _verify_model(self.hub_model_id, "prod", client):
                return bad_val
            if bad_val := _verify_dataset(self.hub_input_dataset_id, "prod", client):
                return bad_val

        # If we have access to a hub deployment, verify we have access to the
        # models in the config for that deployment.
        for (deployment_name, model_id) in self.hub_model_ids_automated.items():
            if client := hub_clients.get(deployment_name, None):
                if bad_val := _verify_model(model_id, deployment_name, client):
                    return bad_val

        # If we have access to a hub deployment, verify we have access to the
        # datasets in the config for that deployment.
        for (
            deployment_name,
            dataset_id,
        ) in self.hub_input_dataset_ids_automated.items():
            if client := hub_clients.get(deployment_name, None):
                if bad_val := _verify_dataset(dataset_id, deployment_name, client):
                    return bad_val

        # If we have access to a hub deployment, verify we have access to the
        # datasets in the config for that deployment.
        for (
            deployment_name,
            dataset_id,
        ) in self.hub_input_channel_last_dataset_ids_automated.items():
            if client := hub_clients.get(deployment_name, None):
                if bad_val := _verify_dataset(dataset_id, deployment_name, client):
                    return bad_val

        return None
