# ---------------------------------------------------------------------
# Copyright (c) 2024 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------
from __future__ import annotations

from typing import cast

from qai_hub import (
    CompileJob,
    Dataset,
    InferenceJob,
    InputSpecs,
    Job,
    JobStatus,
    Model,
    ProfileJob,
)
from qai_hub.client import Client
from qai_hub.hub import _global_client

from qai_hub_models.scorecard.device import ScorecardDevice
from qai_hub_models.scorecard.execution_helpers import (
    for_each_scorecard_path_and_device,
    get_async_job_cache_name,
)
from qai_hub_models.scorecard.path_compile import ScorecardCompilePath
from qai_hub_models.scorecard.path_profile import ScorecardProfilePath
from qai_hub_models.scorecard.results.yaml import (
    CompileScorecardJobYaml,
    InferenceScorecardJobYaml,
    ProfileScorecardJobYaml,
)


def _print_if_not_verbose(client: Client, pstr: str):
    if not client.verbose:
        print(pstr)


def _wait_for_job(job: Job, timeout: int | None) -> JobStatus:
    """Wait for a job and return job status after the wait period."""
    status = job.get_status()
    if not status.failure and not status.success and (not timeout or timeout > 0):
        _print_if_not_verbose(
            job._owner,
            f"Waiting for a max of {timeout or 'inf'} secs for {job.__class__.__name__} {job.job_id} ({job.url})...",
        )

        try:
            return job.wait(timeout)
        except TimeoutError:
            return job.get_status()
    else:
        return status


def _get_successful_compile_job_or_failure_message(
    hub: Client,
    compile_results: CompileScorecardJobYaml,
    compile_job_timeout: int | None,
    model_id: str,
    path: ScorecardCompilePath,
    device: ScorecardDevice,
    cache: dict[str, tuple[CompileJob, JobStatus]],
) -> CompileJob | str:
    compile_job_id = compile_results.get_job_id(
        path, model_id, device, fallback_to_universal_device=True
    )
    if not compile_job_id:
        return "Associated Compile Job Not Found"

    # We cache compile job objects because hub.get_job() can be surprisingly slow sometimes.
    if compile_job_id in cache:
        compile_job, compile_job_status = cache[compile_job_id]
    else:
        compile_job = cast(CompileJob, hub.get_job(compile_job_id))
        compile_job_status = _wait_for_job(compile_job, compile_job_timeout)
        cache[compile_job_id] = compile_job, compile_job_status

    if compile_job_status.success:
        return compile_job
    return f"Compile Job {compile_job.job_id} {compile_job_status.state.name} ({compile_job.url})"


def compile_model(
    model_id: str,
    hub_model: Model | str,
    hub: Client = _global_client,
    input_specs: InputSpecs | None = None,
    model_is_quantized: bool = False,
    output_names: list[str] | None = [],
    channel_first_inputs: list[str] = [],
    channel_first_outputs: list[str] = [],
    results: CompileScorecardJobYaml = CompileScorecardJobYaml(),
) -> CompileScorecardJobYaml:
    if not isinstance(hub_model, Model):
        hub_model = hub.get_model(hub_model)

    output_names_option = None
    if output_names:
        output_names_option = f"--output_names {','.join(output_names)}"

    cl_input_option = None
    if channel_first_inputs:
        cl_input_option = f"--force_channel_last_input {','.join(channel_first_inputs)}"

    cl_output_option = None
    if channel_first_outputs:
        cl_output_option = (
            f"--force_channel_last_output {','.join(channel_first_outputs)}"
        )

    def submit_compile_job(path: ScorecardCompilePath, device: ScorecardDevice):
        job_name = (
            f"qaihm::compile | {get_async_job_cache_name(path, model_id, device)}"
        )

        compile_options = []

        compile_options.append(
            path.get_compile_options(
                model_is_quantized=model_is_quantized,
                device=device.execution_device,
                include_target_runtime=True,
            )
        )

        if output_names_option:
            compile_options.append(output_names_option)

        if path.runtime.channel_last_native_execution:
            if cl_input_option:
                compile_options.append(cl_input_option)
            if cl_output_option:
                compile_options.append(cl_output_option)

        job = cast(
            CompileJob,
            hub.submit_compile_job(
                hub_model,
                device.execution_device,
                job_name,
                input_specs,
                " ".join(compile_options),
            ),
        )
        _print_if_not_verbose(hub, f"{job_name} | Submitted: {job.job_id} | {job.url}")
        results.set_job_id(job.job_id, path, model_id, device)

    for_each_scorecard_path_and_device(
        model_is_quantized, ScorecardCompilePath, submit_compile_job
    )
    return results


def profile_model(
    model_id: str,
    compile_results: CompileScorecardJobYaml,
    hub: Client = _global_client,
    model_is_quantized: bool = False,
    results: ProfileScorecardJobYaml = ProfileScorecardJobYaml(),
    compile_job_timeout: int | None = None,
) -> ProfileScorecardJobYaml:
    compile_job_cache: dict[str, tuple[CompileJob, JobStatus]] = {}

    def submit_profile_job(path: ScorecardProfilePath, device: ScorecardDevice):
        job_name = (
            f"qaihm::profile | {get_async_job_cache_name(path, model_id, device)}"
        )
        compile_job = _get_successful_compile_job_or_failure_message(
            hub,
            compile_results,
            compile_job_timeout,
            model_id,
            path.compile_path,
            device,
            compile_job_cache,
        )
        if isinstance(compile_job, str):
            print(f"{job_name} | Skipped: {compile_job}")
            return

        job = cast(
            ProfileJob,
            hub.submit_profile_job(
                compile_job.get_target_model(),
                device.execution_device,
                job_name,
                path.profile_options,
            ),
        )
        _print_if_not_verbose(hub, f"{job_name} | Submitted: {job.job_id} | {job.url}")
        results.set_job_id(job.job_id, path, model_id, device)

    for_each_scorecard_path_and_device(
        model_is_quantized, ScorecardProfilePath, submit_profile_job
    )
    return results


def inference_model(
    model_id: str,
    hub_dataset: Dataset | str,
    hub_channel_last_dataset: Dataset | str | None,
    compile_results: CompileScorecardJobYaml,
    hub: Client = _global_client,
    model_is_quantized: bool = False,
    results: InferenceScorecardJobYaml = InferenceScorecardJobYaml(),
    compile_job_timeout: int | None = None,
) -> InferenceScorecardJobYaml:
    if not isinstance(hub_dataset, Dataset):
        hub_dataset = hub.get_dataset(hub_dataset)
    if hub_channel_last_dataset and not isinstance(hub_channel_last_dataset, Dataset):
        hub_channel_last_dataset = hub.get_dataset(hub_channel_last_dataset)

    compile_job_cache: dict[str, tuple[CompileJob, JobStatus]] = {}

    def submit_inference_job(path: ScorecardProfilePath, device: ScorecardDevice):
        job_name = (
            f"qaihm::inference | {get_async_job_cache_name(path, model_id, device)}"
        )
        compile_job = _get_successful_compile_job_or_failure_message(
            hub,
            compile_results,
            compile_job_timeout,
            model_id,
            path.compile_path,
            device,
            compile_job_cache,
        )
        if isinstance(compile_job, str):
            print(f"{job_name} | Skipped: {compile_job}")
            return

        dataset = hub_dataset
        if path.runtime.channel_last_native_execution:
            dataset = hub_channel_last_dataset or dataset

        job = cast(
            InferenceJob,
            hub.submit_inference_job(
                compile_job.get_target_model(),
                device.execution_device,
                dataset,
                job_name,
                path.profile_options,
            ),
        )
        _print_if_not_verbose(hub, f"{job_name} | Submitted: {job.job_id} | {job.url}")
        results.set_job_id(job.job_id, path, model_id, device)

    for_each_scorecard_path_and_device(
        model_is_quantized, ScorecardProfilePath, submit_inference_job
    )
    return results
