# ---------------------------------------------------------------------
# Copyright (c) 2024 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------
from __future__ import annotations

import argparse
import datetime
import multiprocessing
import os
from itertools import cycle
from pathlib import Path
from typing import cast

from qai_hub_models.scorecard.internal.model_config import (
    DEFAULT_MODELS_DIR,
    ScorecardModelConfig,
)
from qai_hub_models.scorecard.internal.scripts.sync_model_assets import (
    parse_models_arg_to_model_ids,
)
from qai_hub_models.scorecard.results.spreadsheet import ResultsSpreadsheet
from qai_hub_models.scorecard.results.yaml import (
    CompileScorecardJobYaml,
    InferenceScorecardJobYaml,
    ProfileScorecardJobYaml,
)
from qai_hub_models.utils.path_helpers import SCORECARD_RESULTS_DIR
from qai_hub_models.utils.testing import (
    get_artifacts_dir_opt,
    get_compile_job_ids_file,
    get_inference_job_ids_file,
    get_profile_job_ids_file,
)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--models",
        type=str,
        help="Comma-separated list of models to update. Each entry should map to one of the folders in qai_hub_models/folders.",
        default=os.environ.get("QAIHM_BENCH_TEST_MODELS", "ALL"),
    )
    parser.add_argument(
        "--model-dir",
        type=str,
        help="Model directory in which models can be found.",
        default=os.environ.get("QAIHM_BENCH_TEST_DIR", DEFAULT_MODELS_DIR),
    )
    parser.add_argument(
        "--skip-compile", action="store_true", help="Skip compile job collection."
    )
    parser.add_argument(
        "--skip-profile", action="store_true", help="Skip profile job collection."
    )
    parser.add_argument(
        "--skip-inference", action="store_true", help="Skip inference job collection."
    )
    parser.add_argument(
        "--artifacts-dir",
        type=str,
        metavar="\b",
        default=get_artifacts_dir_opt(),
        help="Directory in which job ID YAMLs are saved.",
    )
    parser.add_argument("--results-dir", type=str, default=SCORECARD_RESULTS_DIR)
    return parser.parse_args()


def process_model(
    model_id: str,
    models_dir: Path,
    compile_job_yamls: CompileScorecardJobYaml | None,
    profile_job_yamls: ProfileScorecardJobYaml | None,
    inference_job_yamls: InferenceScorecardJobYaml | None,
) -> ResultsSpreadsheet:
    def print_with_id(str):
        print(f"{model_id} | {str}")

    # Load config
    model_info = ScorecardModelConfig.from_yaml(models_dir / (model_id + ".yaml"))

    # Get summaries for this model and its components.
    compile_summary = None
    if compile_job_yamls:
        print_with_id("Loading compile summary")
        compile_summary = compile_job_yamls.summary_from_model(
            model_id, model_info.fully_quantized
        )

    profile_summary = None
    if profile_job_yamls:
        print_with_id("Loading profile summary")
        profile_summary = profile_job_yamls.summary_from_model(
            model_id, model_info.fully_quantized
        )

    inference_summary = None
    if inference_job_yamls:
        print_with_id("Loading inference summary")
        inference_summary = inference_job_yamls.summary_from_model(
            model_id, model_info.fully_quantized
        )

    print_with_id("Adding to Spreadsheet")
    entries = ResultsSpreadsheet()
    entries.append_model_summary_entries(
        model_info.fully_quantized, compile_summary, profile_summary, inference_summary
    )
    return entries


if __name__ == "__main__":
    now_str = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")

    args = parse_args()
    model_dir = Path(args.model_dir)
    model_ids = parse_models_arg_to_model_ids(cast(str, args.models), model_dir)
    results_dir = Path(args.results_dir)
    artifacts_dir = Path(args.artifacts_dir) if args.artifacts_dir else None
    compile: bool = not args.skip_compile
    profile: bool = not args.skip_profile
    inference: bool = not args.skip_inference

    if not args.artifacts_dir:
        raise ValueError(
            "Missing artifacts dir. Pass --artifacts-dir or set $ARTIFACTS_DIR"
        )

    if not compile and not profile and not inference:
        print("Please enable compile, profile, or inference")
        exit(1)

    # Make output dir
    os.makedirs(args.results_dir, exist_ok=True)

    # Load YAML
    compile_job_yaml = (
        CompileScorecardJobYaml.from_file(get_compile_job_ids_file(artifacts_dir))
        if compile
        else None
    )
    profile_job_yaml = (
        ProfileScorecardJobYaml.from_file(get_profile_job_ids_file(artifacts_dir))
        if profile
        else None
    )
    inference_job_yaml = (
        InferenceScorecardJobYaml.from_file(get_inference_job_ids_file(artifacts_dir))
        if inference
        else None
    )

    # Extract Data from Models
    if len(model_ids) > 1:
        # Use multiprocessing for multiple models because getting jobs from Hub is slow
        pool = multiprocessing.Pool(processes=15)
        model_spreadsheets = pool.starmap(
            process_model,
            zip(
                model_ids,
                cycle([model_dir]),
                cycle([compile_job_yaml]),
                cycle([profile_job_yaml]),
                cycle([inference_job_yaml]),
            ),
        )
        pool.close()
    else:
        # Single model option for that allows breakpoints
        model_spreadsheets = [
            process_model(
                model_ids[0],
                model_dir,
                compile_job_yaml,
                profile_job_yaml,
                inference_job_yaml,
            )
        ]

    # Collect Model Results
    spreadsheet = ResultsSpreadsheet(compile, profile, inference)
    for model_spreadsheet in model_spreadsheets:
        spreadsheet.extend(model_spreadsheet)

    # Write spreadsheet to disk
    if spreadsheet is not None:
        summary_path = os.path.join(
            args.results_dir, f"internal-scorecard-summary-{now_str}.csv"
        )
        spreadsheet.to_csv(summary_path)
        print(f"Spreadsheet written to {os.path.realpath(summary_path)}")
