# ---------------------------------------------------------------------
# Copyright (c) 2024 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------
from __future__ import annotations

import argparse
import os
from pathlib import Path
from typing import cast

from qai_hub_models.scorecard.internal.hub_clients import get_hub_client_or_raise
from qai_hub_models.scorecard.internal.model_config import (
    DEFAULT_MODELS_DIR,
    ScorecardModelConfig,
)
from qai_hub_models.scorecard.internal.model_exec import (
    compile_model,
    inference_model,
    profile_model,
)
from qai_hub_models.scorecard.internal.model_sync import sync_model_assets
from qai_hub_models.scorecard.internal.scripts.sync_model_assets import (
    parse_models_arg_to_model_ids,
)
from qai_hub_models.scorecard.results.yaml import (
    CompileScorecardJobYaml,
    InferenceScorecardJobYaml,
    ProfileScorecardJobYaml,
)
from qai_hub_models.utils.testing import (
    get_artifacts_dir_opt,
    get_compile_job_ids_file,
    get_inference_job_ids_file,
    get_profile_job_ids_file,
)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "--models",
        type=str,
        metavar="\b",
        help="Comma-separated list of scorecard model IDs to execute.",
        default=os.environ.get("QAIHM_BENCH_TEST_MODELS", "ALL"),
    )
    parser.add_argument(
        "--model-dir",
        type=str,
        metavar="\b",
        help="Directory in which model configs can be found.",
        default=os.environ.get("QAIHM_BENCH_TEST_DIR", DEFAULT_MODELS_DIR),
    )
    parser.add_argument(
        "--skip-compile", action="store_true", help="Skip compile step."
    )
    parser.add_argument(
        "--skip-profile", action="store_true", help="Skip profile step."
    )
    parser.add_argument(
        "--skip-inference", action="store_true", help="Skip inference step."
    )
    parser.add_argument(
        "--compile-ids",
        type=str,
        metavar="\b",
        default=os.environ.get("QAIHM_BENCH_COMPILE_JOBS_FILE", None),
        help="Compile Jobs YAML path. Applicable only if compile step is skipped.",
    )
    parser.add_argument(
        "--compile-timeout",
        type=int,
        metavar="\b",
        default=None,
        help=(
            "Wait a max of this many seconds for *each* compile job to finish before executing a profile or inference job. "
            "If a timeout is reached, the profile / inference job submission for that compiled model is skipped. "
            "If you do not want to wait for compile jobs, this should be set to 0. If you wish to wait indefinitely, "
            "this should be left at the default value."
        ),
    )
    parser.add_argument(
        "--artifacts-dir",
        type=str,
        metavar="\b",
        default=os.environ.get(
            "QAIHM_BENCH_INTERMEDIATES_DIR", get_artifacts_dir_opt()
        ),
        help="Directory in which output files (Job ID YAMLs) should be saved.",
    )
    parser.add_argument(
        "--deployment",
        type=str,
        metavar="\b",
        help="AI Hub deployment to execute jobs on.",
        default=os.environ.get("QAIHM_BENCH_TEST_DEPLOYMENT", "prod"),
    )
    parser.add_argument(
        "--persist-asset-sync",
        action="store_true",
        help=(
            "By default, this script will upload models and temporary datasets to Hub if an asset is not already  "
            "uploaded to the target Hub deployment. These assets are not persisted beyond the lifetime of this script. "
            "If this arg is set, permanent datasets will be synced, and asset IDs will be saved to the on-disk model config. "
            "Generally you should set this only if you anticipate this deployment being used multiple times and you plan to commit the config changes."
        ),
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()

    # Collect args
    models_path: Path = Path(args.model_dir)
    model_id_list = parse_models_arg_to_model_ids(cast(str, args.models), models_path)
    artifacts_dir = Path(args.artifacts_dir)
    os.makedirs(artifacts_dir, exist_ok=True)
    compile_ids = (
        Path(args.compile_ids)
        if args.compile_ids
        else get_compile_job_ids_file(artifacts_dir)
    )
    deployment: str = args.deployment
    compile_job_timeout: int | None = args.compile_timeout
    permanent_asset_sync: bool = args.persist_asset_sync

    compile: bool = not args.skip_compile
    profile: bool = not args.skip_profile
    inference: bool = not args.skip_inference
    if not compile and not profile and not inference:
        print("Please enable compile, profile, or inference")
        exit(1)

    # Get model configs
    model_configs = [
        ScorecardModelConfig.from_scorecard_model_id(model_id)
        for model_id in model_id_list
    ]

    # Sync assets for selected models
    for config in model_configs:
        print(f"Sync Assets for {config.id}")
        sync_model_assets(config, deployment, permanent_asset_sync)
        if permanent_asset_sync:
            config.to_scorecard_yaml(models_path)

    # Disable client verbosity; compile/profile/inference model will print more succinct summaries instead
    hub = get_hub_client_or_raise(deployment)
    hub.set_verbose(False)

    compile_results = None
    if compile:
        compile_results = CompileScorecardJobYaml()
        for config in model_configs:
            compile_model(
                config.id,
                config.hub_model_ids_automated[deployment],
                hub,
                config.get_hub_api_input_specs(),
                config.fully_quantized,
                config.output_names,
                config.channel_first_inputs,
                config.channel_first_outputs,
                compile_results,
            )
        compile_results.to_file(get_compile_job_ids_file(artifacts_dir))
    elif profile or inference:
        compile_results = CompileScorecardJobYaml.from_file(args.compile_ids)

    if profile:
        assert compile_results  # for mypy
        profile_results = ProfileScorecardJobYaml()
        for config in model_configs:
            profile_model(
                config.id,
                compile_results,
                hub,
                config.fully_quantized,
                profile_results,
                compile_job_timeout,
            )
        profile_results.to_file(get_profile_job_ids_file(artifacts_dir))

    if inference:
        assert compile_results  # for mypy
        inference_results = InferenceScorecardJobYaml()
        for config in model_configs:
            inference_model(
                config.id,
                config.hub_input_dataset_ids_automated[deployment],
                config.hub_input_channel_last_dataset_ids_automated.get(deployment),
                compile_results,
                hub,
                config.fully_quantized,
                inference_results,
                compile_job_timeout,
            )
        inference_results.to_file(get_inference_job_ids_file(artifacts_dir))
