# ---------------------------------------------------------------------
# Copyright (c) 2024 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------
import argparse
import os
from pathlib import Path
from typing import cast

from qai_hub_models.scorecard.internal.model_config import (
    DEFAULT_MODELS_DIR,
    ScorecardModelConfig,
)
from qai_hub_models.scorecard.internal.model_sync import (
    DEFAULT_DEPLOYMENTS,
    sync_model_assets,
)


def parse_models_arg_to_model_ids(models: str, models_path: Path) -> list[str]:
    if models == "ALL":
        model_id_list = [
            x[:-5]
            for x in os.listdir(models_path)
            if x.endswith(".yaml") and not x.endswith("_example.yaml")
        ]
    else:
        model_id_list = [x.strip() for x in models.split(",")]

    return model_id_list


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--models",
        type=str,
        help="Comma-separated list of models (scorecard model IDs) to sync.",
        default=os.environ.get("QAIHM_BENCH_TEST_MODELS", "ALL"),
    )
    parser.add_argument(
        "--model-dir",
        type=str,
        help="Model directory in which models can be found.",
        default=os.environ.get("QAIHM_BENCH_TEST_DIR", DEFAULT_MODELS_DIR),
    )
    parser.add_argument(
        "--deployments",
        type=str,
        help="Comma-separated list of hub deployment (names) to sync.",
        default=os.environ.get(
            "QAIHM_BENCH_TEST_DEPLOYMENTS", ",".join(DEFAULT_DEPLOYMENTS)
        ),
    )
    parser.add_argument(
        "--clear-existing-assets",
        action="store_true",
        help="Re-uploads all assets to all deployments, regardless of whether they exist already in a config or not.",
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()

    models_path = Path(cast(str, args.model_dir))
    model_id_list = parse_models_arg_to_model_ids(cast(str, args.models), models_path)
    deployments = cast(str, args.deployments)
    deployments_list = {x.strip() for x in deployments.split(",")}
    for model_id in model_id_list:
        config = ScorecardModelConfig.from_scorecard_model_id(model_id)
        sync_model_assets(
            config,
            deployments_list,
            permanent_dataset_upload=True,
            clear_existing=args.clear_existing_assets,
        )
        config.to_scorecard_yaml(models_path)
