# ---------------------------------------------------------------------
# Copyright (c) 2024 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------
import argparse
import copy
import importlib
import os
import subprocess
import traceback

import qai_hub as hub
from ruamel.yaml import YAML

from qai_hub_models.models.common import ExportResult
from qai_hub_models.utils.asset_loaders import qaihm_temp_dir
from qai_hub_models.utils.base_model import TargetRuntime
from qai_hub_models.utils.measurement import (
    get_checkpoint_file_size,
    get_tflite_unique_parameters,
)
from qai_hub_models.utils.path_helpers import MODEL_IDS, QAIHM_MODELS_ROOT


def get_model_size_and_parameters(job):
    if isinstance(job, hub.ProfileJob) or isinstance(job, hub.InferenceJob):
        with qaihm_temp_dir() as tmp_dirname:
            path = os.path.join(tmp_dirname, "model.tflite")
            job.model.download(path)
            size = get_checkpoint_file_size(path)
            parameters = get_tflite_unique_parameters(path)
            return size, parameters
    else:
        with qaihm_temp_dir() as tmp_dirname:
            path = os.path.join(tmp_dirname, "model.tflite")
            job.download_target_model(path)
            size = get_checkpoint_file_size(path)
            parameters = get_tflite_unique_parameters(path)
            return size, parameters


def add_details_to_info_yaml(info, details):
    # Clean the old keys
    new_info = copy.deepcopy(info)
    if "technical_details" in info:
        for info_type in info["technical_details"]:
            if "model size" in info_type.lower() or "parameters" in info_type.lower():
                del new_info["technical_details"][info_type]

    # Add more standard keys here
    for name, (size, parameters) in details.items():
        if "technical_details" in new_info:

            param_key = (
                "Number of parameters"
                if name == "model"
                else f"Number of parameters ({name})"
            )
            size_key = "Model size" if name == "model" else f"Model size ({name})"
            new_info["technical_details"][param_key] = parameters
            new_info["technical_details"][size_key] = size

    return new_info


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--models",
        "-m",
        nargs="+",
        type=str,
        default=None,
        help="Models for which to autofill info.yaml.",
    )
    parser.add_argument(
        "--all",
        "-a",
        action="store_true",
        help="If set, generates files for all models.",
    )
    parser.add_argument(
        "--skip-compile-jobs",
        "-s",
        action="store_true",
        help="Skip running compile jobs.",
        default=False,
    )
    args = parser.parse_args()
    skip_compile = args.skip_compile_jobs

    assert args.all or args.models, "Must specify -a or -m."
    if args.models:
        models = parser.parse_args().models
    else:
        models = MODEL_IDS

    for model_name in models:
        model_dir = QAIHM_MODELS_ROOT / model_name
        yaml = YAML()
        yaml.indent(mapping=2, sequence=2, offset=2)
        export_options = {}
        if (model_dir / "code-gen.yaml").exists():
            with open(model_dir / "code-gen.yaml") as f:
                export_options = yaml.load(f)
        if (model_dir / "perf.yaml").exists():
            with open(model_dir / "perf.yaml") as f:
                perf = yaml.load(f)

        details = dict()

        if export_options.get("is_precompiled", False):
            print(f"Skipping {model_name} since its a is_precompiled asset.")

            exit(1)
        try:
            if not skip_compile:

                # Install dependencies
                requirements_file = None

                if (model_dir / "requirements.txt").exists():
                    requirements_file = os.path.join(model_dir, "requirements.txt")
                if requirements_file:
                    subprocess.run(["pip", "install", "-r", requirements_file])

                # imports the module from the given path
                model_module = importlib.import_module(
                    f"qai_hub_models.models.{model_name}.export"
                )
                results = model_module.export_model(
                    device="Samsung Galaxy S23 (Family)",
                    skip_downloading=True,
                    skip_profiling=True,
                    skip_inferencing=True,
                    skip_summary=True,
                    target_runtime=TargetRuntime.TFLITE,
                )

                if isinstance(results, ExportResult):
                    details["model"] = get_model_size_and_parameters(
                        results.compile_job
                    )
                else:
                    for component_name in results:
                        details[component_name] = get_model_size_and_parameters(
                            results[component_name].compile_job
                        )

            else:

                if "components" in export_options:
                    for m in perf["models"]:
                        for submodel_name, submodel in export_options[
                            "components"
                        ].items():
                            if m["name"] == submodel_name:
                                job_id = m["performance_metrics"][0][
                                    "torchscript_onnx_tflite"
                                ]["job_id"]
                                details[submodel_name] = get_model_size_and_parameters(
                                    hub.get_job(job_id)
                                )

                else:
                    job_id = perf["models"][0]["performance_metrics"][0][
                        "torchscript_onnx_tflite"
                    ]["job_id"]
                    details["model"] = get_model_size_and_parameters(
                        hub.get_job(job_id)
                    )

            if (model_dir / "info.yaml").exists():
                with open(model_dir / "info.yaml") as f:
                    info = yaml.load(f)
                    new_info = add_details_to_info_yaml(info=info, details=details)
                    with open(model_dir / "info.yaml", "wb") as f:
                        yaml.dump(new_info, f)
            else:
                new_info = add_details_to_info_yaml(info=dict(), details=details)
                with open(model_dir / "info.yaml", "w") as f:
                    yaml.dump(new_info, f)

        except Exception:
            print(traceback.format_exc())
            print(f"Not working {model_name}")


if __name__ == "__main__":
    main()
