# ---------------------------------------------------------------------
# Copyright (c) 2024 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------
# Copyright (c) 2024 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------
# ---------------------------------------------------------------------
import os
import tempfile

import boto3
from botocore.client import BaseClient
from tqdm import tqdm

from qai_hub_models.datasets.cityscapes import (
    GT_DIR_NAME,
    IMAGES_DIR_NAME,
    CityscapesDataset,
)
from qai_hub_models.utils.asset_loaders import LOCAL_STORE_DEFAULT_PATH

AWS_PROFILE_VAR = "AWS_PROFILE"
S3_BUCKET = "qai-hub-models-private-assets"


class S3ClientSingleton:
    _instance = None

    def __new__(cls) -> BaseClient:
        if cls._instance is None:
            cls._instance = boto3.client("s3")
        return cls._instance


class AWSEnvManager:
    def __enter__(self):
        self.profile = os.environ.get(AWS_PROFILE_VAR, None)
        os.environ[AWS_PROFILE_VAR] = "tetra-biz-data"

    def __exit__(self, exc_type, exc_value, traceback):
        if self.profile is not None:
            os.environ[AWS_PROFILE_VAR] = "tetra-biz-data"


def download_file(object_key: str, filepath: str):
    s3_client = S3ClientSingleton()
    assert isinstance(s3_client, BaseClient)
    meta_data = s3_client.head_object(Bucket=S3_BUCKET, Key=object_key)
    total_length = int(meta_data.get("ContentLength", 0))

    with tqdm(total=total_length, unit="B", unit_scale=True) as t:
        s3_client.download_file(S3_BUCKET, object_key, filepath, Callback=t.update)


def main() -> None:
    try:
        # If loading succeeds, the data is already cached
        CityscapesDataset()
    except ValueError:
        os.makedirs(LOCAL_STORE_DEFAULT_PATH, exist_ok=True)
        with tempfile.TemporaryDirectory(dir=LOCAL_STORE_DEFAULT_PATH) as tmp_dir:
            _, gt_zip = tempfile.mkstemp(suffix=f"{GT_DIR_NAME}.zip", dir=tmp_dir)
            _, images_zip = tempfile.mkstemp(
                suffix=f"{IMAGES_DIR_NAME}.zip", dir=tmp_dir
            )
            download_file(
                f"qai-hub-models/datasets/cityscapes/{IMAGES_DIR_NAME}.zip", images_zip
            )
            download_file(
                f"qai-hub-models/datasets/cityscapes/{GT_DIR_NAME}.zip", gt_zip
            )

            # Initialize the dataset, which will extract data to disk
            CityscapesDataset(input_images_zip=images_zip, input_gt_zip=gt_zip)


if __name__ == "__main__":
    with AWSEnvManager():
        main()
