# ---------------------------------------------------------------------
# Copyright (c) 2024 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------
from __future__ import annotations

import argparse
import os
import subprocess
from pathlib import Path
from typing import Any

from jinja2 import Environment, FileSystemLoader

from qai_hub_models.configs.info_yaml import QAIHMModelCodeGen, QAIHMModelInfo
from qai_hub_models.models.common import TargetRuntime
from qai_hub_models.utils.path_helpers import MODEL_IDS, QAIHM_MODELS_ROOT

HEADER = "# THIS FILE WAS AUTO-GENERATED. DO NOT EDIT MANUALLY."

# Empirically measured limit of number of samples per inference for each dataset
DEFAULT_SPLIT_SIZES = {
    "imagenette": 2500,
    "imagenet": 2500,
    "pascal_voc": 400,
    "bsd300": 100,
    "coco": 250,
    "nyuv2": 100,
    "nyuv2x518": 100,
    "cityscapes": 50,
}


def _skip_clone_repo_check(model_dir):
    original_test_path = model_dir / "test.py"
    if not original_test_path.exists():
        return False
    with open(original_test_path) as f:
        original_file_contents = f.read()
    return "skip_clone_repo_check" in original_file_contents


def _get_steps(export_options: QAIHMModelCodeGen) -> dict[str, str]:
    steps = {}
    # step_number needs to be an object (not a primitive) for its state
    # to persist across function calls
    step_number = 1

    def add_step(name, step):
        nonlocal step_number
        steps[name] = str(step_number) + ". " + step
        step_number += 1

    if export_options.is_precompiled:
        add_step("init", "Initialize model")
        add_step("upload", "Upload model assets to hub")
    else:
        add_step(
            "init",
            "Instantiates a PyTorch model and converts it to a traced TorchScript format",
        )
        if export_options.use_hub_quantization:
            add_step(
                "quantize",
                "Converts the PyTorch model to ONNX and quantizes the ONNX model.",
            )
        add_step("compile", "Compiles the model to an asset that can be run on device")
    add_step("profile", "Profiles the model performance on a real device")

    if not export_options.is_precompiled:
        add_step("inference", "Inferences the model on sample inputs")
        add_step("download", "Downloads the model asset to the local directory")

    summary_step = "Summarizes the results from profiling"
    if not export_options.is_precompiled:
        summary_step += " and inference"
    add_step("summary", summary_step)
    return steps


def _extract_runtime_options(
    export_options: QAIHMModelCodeGen, export_options_dict: dict | None = None
) -> dict[str, Any]:
    export_options_dict = export_options_dict or {}
    default_runtime = None
    for runtime in TargetRuntime:
        runtime_support = export_options.supports_runtime(runtime)
        export_options_dict[f"supports_{runtime.name.lower()}"] = runtime_support
        if not default_runtime and runtime_support:
            default_runtime = runtime.name.upper()

    if not default_runtime:
        # if all runtimes are disabled, just pick one so the generated code is valid. The model can't be public in this case.
        default_runtime = "TFLITE"

    export_options_dict["default_runtime"] = default_runtime
    return export_options_dict


def _generate_export(
    environment,
    model_name,
    model_display_name,
    export_options: QAIHMModelCodeGen,
    model_dir,
):
    template = environment.get_template("export_template.txt")

    export_options_dict = export_options.to_dict()
    _extract_runtime_options(export_options, export_options_dict)

    file_contents = template.render(
        export_options_dict,
        model_name=model_name,
        model_display_name=model_display_name,
        header=HEADER,
        steps=_get_steps(export_options),
    )
    export_file_path = os.path.join(model_dir, "export.py")
    with open(export_file_path, "w") as f:
        f.write(file_contents)
    return export_file_path


def _generate_unit_tests(environment, model_name, export_options, test_path) -> str:
    template = environment.get_template("unit_test_template.txt")
    if export_options["eval_datasets"]:
        default_split_size = DEFAULT_SPLIT_SIZES[export_options["eval_datasets"][-1]]
        num_eval_samples = min(default_split_size, 100)
    else:
        num_eval_samples = 100
    file_contents = template.render(
        export_options,
        model_name=model_name,
        header=HEADER,
        num_eval_samples=num_eval_samples,
    )

    with open(test_path, "w") as f:
        f.write(file_contents)
    return str(test_path)


def _generate_conftest(environment, model_name, export_options, conftest_path) -> str:
    template = environment.get_template("conftest_template.txt")
    file_contents = template.render(
        export_options,
        model_name=model_name,
        header=HEADER,
    )
    with open(conftest_path, "w") as f:
        f.write(file_contents)
    return str(conftest_path)


def _generate_evaluate(environment, model_name, export_options, evaluate_path):
    template = environment.get_template("evaluate_template.txt")
    file_contents = template.render(
        export_options,
        model_name=model_name,
        header=HEADER,
        default_split_size=DEFAULT_SPLIT_SIZES[export_options["eval_datasets"][0]],
    )
    with open(evaluate_path, "w") as f:
        f.write(file_contents)
    return str(evaluate_path)


def generate_export_file(model_name: str) -> list[str]:
    model_dir = QAIHM_MODELS_ROOT / model_name
    export_options = QAIHMModelCodeGen.from_yaml(model_dir / "code-gen.yaml")

    if export_options.skip_export:
        print(f"Skipping export.py generation for {model_name}.")
        return []

    try:
        model_info = QAIHMModelInfo.from_model(model_name)
        model_display_name = model_info.name
    except ValueError:
        # Info yaml does not exist
        model_display_name = "no_info_yaml_found"

    environment = Environment(
        loader=FileSystemLoader(Path(__file__).parent / "templates/"),
        keep_trailing_newline=True,
    )
    generated_files = [
        _generate_export(
            environment, model_name, model_display_name, export_options, model_dir
        )
    ]
    export_options_dict = export_options.to_dict()
    _extract_runtime_options(export_options, export_options_dict)

    should_generate_tests = not export_options.skip_hub_tests_and_scorecard
    test_path = model_dir / "test_generated.py"
    if should_generate_tests:
        export_options_dict["skip_clone_repo"] = _skip_clone_repo_check(model_dir)
        generated_files.append(
            _generate_unit_tests(
                environment, model_name, export_options_dict, test_path
            )
        )
    elif test_path.exists():
        os.remove(test_path)

    should_generate_conftest = (
        should_generate_tests and not export_options.is_precompiled
    )
    conftest_path = model_dir / "conftest.py"
    if should_generate_conftest:
        generated_files.append(
            _generate_conftest(
                environment, model_name, export_options_dict, conftest_path
            )
        )
    elif conftest_path.exists():
        os.remove(conftest_path)

    evaluate_path = model_dir / "evaluate.py"
    if export_options.eval_datasets:
        generated_files.append(
            _generate_evaluate(
                environment, model_name, export_options_dict, evaluate_path
            )
        )
    elif evaluate_path.exists():
        os.remove(evaluate_path)

    zoo_root = QAIHM_MODELS_ROOT
    return [os.path.join(zoo_root, gen_file) for gen_file in generated_files]


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--models",
        "-m",
        nargs="+",
        type=str,
        default=None,
        help="Models for which to generate export.py.",
    )
    parser.add_argument(
        "--all",
        "-a",
        action="store_true",
        help="If set, generates files for all models.",
    )
    args = parser.parse_args()
    assert args.all or args.models, "Must specify -a or -m."
    if args.models:
        models = parser.parse_args().models
    else:
        models = MODEL_IDS
    modified_files = []
    for model in models:
        modified_files.extend(generate_export_file(model))
    os.environ["SKIP"] = "mypy"
    subprocess.run(["pre-commit", "run", "--files"] + modified_files)


if __name__ == "__main__":
    main()
