# ---------------------------------------------------------------------
# Copyright (c) 2024 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------
import argparse

import qai_hub as hub

from qai_hub_models.scorecard.device import cs_8_elite, cs_8_gen_3, cs_x_elite
from qai_hub_models.scorecard.results.performance_summary import (
    get_reference_device_info,
)
from qai_hub_models.scripts.generate_perf_yaml import write_jobs_config


def main():
    parser = argparse.ArgumentParser(
        description="Create a YAML file with specified keys."
    )
    parser.add_argument(
        "--supports-gen-8-elite",
        type=bool,
        default=True,
        help="Pass the flag if 8 Elite is supported",
    )
    parser.add_argument(
        "--supports-gen-8-gen-3",
        type=bool,
        default=False,
        help="Pass the flag if 8 Gen 3 is supported",
    )
    parser.add_argument(
        "--supports-gen-8-gen-2",
        type=bool,
        default=False,
        help="Pass the flag if 8 Gen 3 is supported",
    )
    parser.add_argument(
        "--model", help="Model Name. Must match the folder name.", required=True
    )
    args = parser.parse_args()

    supported_devices = []
    supported_chipsets = []
    ref_devices = []
    if args.supports_gen_8_elite:
        supported_devices.extend(
            [
                dev.name
                for dev in hub.get_devices(
                    attributes="chipset:qualcomm-snapdragon-8-elite"
                )
            ]
        )
        supported_chipsets.append("Snapdragon® 8 Elite")
        ref_devices.append(get_reference_device_info(cs_8_elite))
    if args.supports_gen_8_gen_3:
        supported_devices.extend(
            [
                dev.name
                for dev in hub.get_devices(
                    attributes="chipset:qualcomm-snapdragon-8gen3"
                )
            ]
        )
        supported_chipsets.append("Snapdragon® 8 Gen 3")
        ref_devices.append(get_reference_device_info(cs_8_gen_3))
    if args.supports_gen_8_gen_2:
        supported_devices.extend(
            [
                dev.name
                for dev in hub.get_devices(
                    attributes="chipset:qualcomm-snapdragon-8gen2"
                )
            ]
        )
        supported_chipsets.extend(["Snapdragon® 8 Gen 2", "Snapdragon® X Elite"])
        ref_devices.append(get_reference_device_info(cs_x_elite))

    perf_dict = dict(
        aggregated=dict(
            supported_oses=["Android"],
            supported_devices=supported_devices,
            supported_chipsets=supported_chipsets,
        )
    )
    for ref_device in ref_devices:
        metrics = dict(
            llm_metrics=dict(
                time_to_first_token_range=dict(min="", max=""), tokens_per_second=""
            ),
            evaluation_metrics=dict(name="", value="", unit=""),
            inference_time="null",
            throughput="null",
            estimated_peak_memory_range=dict(min="null", max="null"),
            primary_compute_unit="null",
            precision="null",
            layer_info=dict(
                layers_on_npu="null", layers_on_gpu="null", layers_on_cpu="null"
            ),
        )
        performance_metrics = [
            dict(torchscript_onnx_qnn=metrics, reference_device_info=ref_device)
        ]
    perf_dict["models"] = dict(name="", performance_metrics=performance_metrics)
    write_jobs_config(perf_dict, f"qai_hub_models/models/{args.model}/perf.yaml")

    print(f"Created perf.yaml for model {args.model} successfully.")


if __name__ == "__main__":
    main()
