# ---------------------------------------------------------------------
# Copyright (c) 2024 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------
import argparse
import os
import subprocess
import time
from typing import Optional

import qai_hub as hub
import ruamel.yaml
from huggingface_hub import create_repo, upload_file

from qai_hub_models.configs._info_yaml_llm_details import LLM_CALL_TO_ACTION
from qai_hub_models.configs.info_yaml import MODEL_STATUS, QAIHMModelInfo
from qai_hub_models.configs.perf_yaml import QAIHMModelPerf, bytes_to_mb
from qai_hub_models.models.common import TargetRuntime
from qai_hub_models.scorecard.device import cs_universal
from qai_hub_models.utils.asset_loaders import (
    ASSET_BASES_DEFAULT_PATH,
    ASSET_CONFIG,
    QAIHM_WEB_ASSET,
    ModelZooAssetConfig,
    load_yaml,
    qaihm_temp_dir,
)
from qai_hub_models.utils.path_helpers import MODEL_IDS
from qai_hub_models.utils.printing import _INFO_DASH, get_profile_metrics
from qai_hub_models.utils.readme_util import (
    _get_licenses,
    _get_package_instructions,
    _get_references,
    _get_usage_and_limitations,
)

AIHUB_URL = "https://aihub.qualcomm.com"
ORG_NAME = "qualcomm"

MODEL_CARD_TEMPLATE = open(
    os.path.join(os.path.dirname(__file__), "templates", "hf_model_card_template.md")
).read()

NOASSSETS_MODEL_CARD_TEMPLATE = open(
    os.path.join(
        os.path.dirname(__file__), "templates", "hf_noassets_model_card_template.md"
    )
).read()


PACKAGE_INSTALL_INSTRUCTIONS = """
Install the package via pip:
```bash
pip install qai-hub-models
```
"""


CONFIGURE_INSTRUCTIONS = """
Sign-in to [Qualcomm® AI Hub](https://app.aihub.qualcomm.com/) with your
Qualcomm® ID. Once signed in navigate to `Account -> Settings -> API Token`.

With this API token, you can configure your client to run models on the cloud
hosted devices.
```bash
qai-hub configure --api_token API_TOKEN
```
Navigate to [docs](https://app.aihub.qualcomm.com/docs/) for more information.
"""


def upload_hub_asset(
    num_retries: int,
    profile_job: hub.ProfileJob,
    model_path: str,
    name: str,
    model_id: str,
    is_precompiled=False,
) -> None:
    if num_retries == 0:
        return
    try:
        if is_precompiled:
            model_path = profile_job.model.download(model_path)  # type: ignore
        else:
            profile_job_model_producer = profile_job.model.producer
            assert isinstance(profile_job_model_producer, hub.CompileJob)
            model_path = profile_job_model_producer.download_target_model(model_path)  # type: ignore

        upload_file(
            path_or_fileobj=model_path,
            path_in_repo=os.path.basename(model_path),
            repo_id=f"qualcomm/{model_id}",
        )
    except Exception:
        time.sleep(60)
        return upload_hub_asset(
            num_retries - 1,
            profile_job,
            model_path,
            name,
            model_id,
            is_precompiled,
        )


def upload_readme_to_hf(
    private_status: bool,
    model_card_path: str,
    model_name: str,
    repo_id: str,
):
    create_repo(repo_id=repo_id, exist_ok=True, private=private_status)
    upload_file(
        path_or_fileobj=model_card_path,
        path_in_repo="README.md",
        repo_id=f"qualcomm/{model_name}",
    )


def main():
    """Generate a hugging face model card for each model, and save it in <model_dir>/HF_MODEL_CARD.md."""
    args = parse_args()

    # Login using HuggingFace CLI
    subprocess.run(["huggingface-cli", "login", "--token", args.huggingface_token])

    # Verify correct environment variables are set.
    ModelZooAssetConfig.load_asset_cfg(
        ASSET_BASES_DEFAULT_PATH, verify_env_has_all_variables=True
    )

    # Make a list of models, we wish to upload to HuggingFace
    assert args.all or args.models
    if args.models:
        model_list = args.models.split(",")
    else:
        model_list = MODEL_IDS

    for model_id in model_list:
        model_info = QAIHMModelInfo.from_model(model_id)
        model_perf = QAIHMModelPerf(
            perf_yaml_path=model_info.get_perf_yaml_path(), model_name=model_info.name
        )
        model_code_gen = {}
        model_code_gen_path = model_info.get_code_gen_yaml_path()
        if os.path.exists(model_code_gen_path):
            model_code_gen = load_yaml(model_code_gen_path)

        repo_id = ORG_NAME + "/" + model_info.name
        private = True if not args.public else False

        # Generate Proprietary model card for
        # models with no code or model asset to share
        # but want to show perf data.
        if model_info.status == MODEL_STATUS.PUBLIC:
            use_noasset_template = (
                model_code_gen.get("skip_example_usage", False)
                or model_info.restrict_model_sharing
            )

            print(f"Creating hugging face model card for {model_id}")
            model_card_path = model_info.get_hf_model_card_path()
            model_card = generate_hf_model_card(
                model_info,
                model_perf,
                model_code_gen,
                model_info.has_model_requirements(),
                NOASSSETS_MODEL_CARD_TEMPLATE
                if use_noasset_template
                else MODEL_CARD_TEMPLATE,
                add_export_out=not use_noasset_template,
            )

            if os.path.exists(model_card_path):
                os.remove(model_card_path)
            with open(model_card_path, "w") as model_card_file:
                model_card_file.write(model_card)

            upload_readme_to_hf(
                private_status=private,
                model_card_path=str(model_card_path),
                model_name=model_info.name,
                repo_id=repo_id,
            )

            # Skip model upload if globally disabled via argument or for specific model
            if args.skip_upload_assets or model_info.restrict_model_sharing:
                continue

            with qaihm_temp_dir() as tmp_dir:
                for model_name, model_perf in model_perf.per_model_details.items():
                    # Try the default scorecard device first.
                    device_perf = model_perf.details_per_device.get(
                        cs_universal.reference_device.name, None
                    )
                    # If that doesn't work, just use the first one in the list.
                    if not device_perf:
                        device_perf = next(iter(model_perf.details_per_device.values()))

                    for (
                        runtime,
                        runtime_perf,
                    ) in device_perf.details_per_path.items():
                        perf_details = runtime_perf.perf_details
                        if isinstance(perf_details, QAIHMModelPerf.PerformanceDetails):
                            profile_job = hub.get_job(perf_details.job_id)
                            assert isinstance(profile_job, hub.ProfileJob)
                            if profile_job.get_status().success:
                                upload_hub_asset(
                                    num_retries=5,
                                    profile_job=profile_job,
                                    model_path=f"{tmp_dir}/{model_name}",  # Hub will add an extension for us.
                                    name=model_name,
                                    model_id=model_info.name,
                                )
                            else:
                                print(
                                    f"{model_name}'s profile job for target {runtime.long_name} failed."
                                )
        else:
            print(
                f"{model_id} was changed from public to private. Manually delete the model if needed otherwise check what happened and run this script with --model {model_id}"
            )


def _get_compiled_model_upload_section(model_package: str, model_components: dict):
    each_model_upload = ""
    for name, obj in model_components.items():
        each_model_upload += f"""
model_{name} = hub.upload_model({obj}.get_target_model_path())"""

    return f"""
Step 1: **Upload compiled model**

Upload compiled models from `{model_package}` on hub.
```python
import torch

import qai_hub as hub
from {model_package} import Model

# Load the model
model = Model.from_precompiled()
{each_model_upload}
```"""


def _get_compilation_section(model_package: str):
    return f"""
Step 1: **Compile model for on-device deployment**

To compile a PyTorch model for on-device deployment, we first trace the model
in memory using the `jit.trace` and then call the `submit_compile_job` API.

```python
import torch

import qai_hub as hub
from {model_package} import Model

# Load the model
torch_model = Model.from_pretrained()

# Device
device = hub.Device("Samsung Galaxy S24")

# Trace model
input_shape = torch_model.get_input_spec()
sample_inputs = torch_model.sample_inputs()

pt_model = torch.jit.trace(torch_model, [torch.tensor(data[0]) for _, data in sample_inputs.items()])

# Compile model on a specific device
compile_job = hub.submit_compile_job(
    model=pt_model,
    device=device,
    input_specs=torch_model.get_input_spec(),
)

# Get target model to run on-device
target_model = compile_job.get_target_model()

```"""


def _get_compilation_section_with_components(model_package, components):
    load_commands = "# Load the model"
    load_commands += "\nmodel = Model.from_pretrained()"
    compile_commands = ""

    for comp_expr in components.values():
        comp_name = comp_expr.split(".")[1]
        load_commands += f"""
{comp_name}_model = {comp_expr}"""
        compile_commands += f"""# Trace model
{comp_name}_input_shape = {comp_name}_model.get_input_spec()
{comp_name}_sample_inputs = {comp_name}_model.sample_inputs()

traced_{comp_name}_model = torch.jit.trace({comp_name}_model, [torch.tensor(data[0]) for _, data in {comp_name}_sample_inputs.items()])

# Compile model on a specific device
{comp_name}_compile_job = hub.submit_compile_job(
    model=traced_{comp_name}_model ,
    device=device,
    input_specs={comp_name}_model.get_input_spec(),
)

# Get target model to run on-device
{comp_name}_target_model = {comp_name}_compile_job.get_target_model()
"""

    return f"""
Step 1: **Compile model for on-device deployment**

To compile a PyTorch model for on-device deployment, we first trace the model
in memory using the `jit.trace` and then call the `submit_compile_job` API.

```python
import torch

import qai_hub as hub
from {model_package} import Model

{load_commands}

# Device
device = hub.Device("Samsung Galaxy S23")

{compile_commands}
```"""


def _get_profile_models_section(
    is_precompiled: bool, precompiled_model_components: dict, components: dict
):
    if not is_precompiled:
        if components == {}:
            return """profile_job = hub.submit_profile_job(
    model=target_model,
    device=device,
)
        """
        else:
            profile_command = ""
            for name in components:
                var = components[name].split(".")[1]
                profile_command += f"""{var}_profile_job = hub.submit_profile_job(
    model={var}_target_model,
    device=device,
)
"""
            return profile_command

    all_model_profile = """
# Device
device = hub.Device("Samsung Galaxy S23")
"""

    for name in precompiled_model_components.keys():
        all_model_profile += f"""profile_job_{name} = hub.submit_profile_job(
    model=model_{name},
    device=device,
)
"""
    return all_model_profile


def _get_inference_models_section(
    is_precompiled: bool, precompiled_model_components: dict, components: dict
):
    if not is_precompiled:
        if components == {}:
            return """input_data = torch_model.sample_inputs()
inference_job = hub.submit_inference_job(
    model=target_model,
    device=device,
    inputs=input_data,
)
    on_device_output = inference_job.download_output_data()
"""
        else:
            ondevice_command = ""
            for name in components:
                var = components[name].split(".")[1]
                ondevice_command += f"""{var}_input_data = {var}_model.sample_inputs()
{var}_inference_job = hub.submit_inference_job(
    model={var}_target_model,
    device=device,
    inputs={var}_input_data,
)
{var}_inference_job.download_output_data()
"""

            return ondevice_command

    all_model_inference = ""
    for name, obj in precompiled_model_components.items():
        all_model_inference += f"""
input_data_{name} = {obj}.sample_inputs()
inference_job_{name} = hub.submit_inference_job(
    model=model_{name},
    device=device,
    inputs=input_data_{name},
)
on_device_output_{name} = inference_job_{name}.download_output_data()
"""
    return all_model_inference


def _get_profile_model_intro(is_precompiled_model):
    model_state = "uploading compiled" if is_precompiled_model else "compiling"
    return f"After {model_state} models from step 1"


def _get_load_model_section(model_package):
    return f"""
### Load the model

Load the model:

```python
from {model_package} import Model
model = Model.from_pretrained()
```"""


def _get_perf_summary(model_perf: QAIHMModelPerf):
    out = "```\nProfiling Results"
    for component, model_perf_detail in model_perf.per_model_details.items():
        out += f"\n{_INFO_DASH}\n{component}"

        # Default to "universal" (default) device if it's there. Otherwise, just use the first one the iter gives us.
        device_perf = model_perf_detail.details_per_device.get(cs_universal.name, None)
        device_perf = device_perf or next(
            iter(model_perf_detail.details_per_device.values())
        )

        # Pick one runtime randomly (the first the iter gives us)
        runtime_perf = next(iter(device_perf.details_per_path.values()))
        out += (
            "\n"
            + get_profile_metrics(
                device_perf.device,
                runtime_perf.path.runtime,
                runtime_perf.perf_details,
            )
            + "\n"
        )

    return out + "```"


def _get_deployment_instructions(is_precompiled):
    formats = "`.so` export " if not is_precompiled else " `.so` / `.bin` export "
    format_desc = (
        "`.so` shared library "
        if not is_precompiled
        else "`.so` shared library or `.bin` context binary"
    )
    return f"""
The models can be deployed using multiple runtimes:
- TensorFlow Lite (`.tflite` export): [This
  tutorial](https://www.tensorflow.org/lite/android/quickstart) provides a
  guide to deploy the .tflite model in an Android application.\n\n
- QNN ({formats}): This [sample
  app](https://docs.qualcomm.com/bundle/publicresource/topics/80-63442-50/sample_app.html)
provides instructions on how to use the {format_desc} in an Android application.
"""


def _get_demo_section(is_precompiled, model_package):
    headline = "Demo on-device" if is_precompiled else "Demo off target"
    return f"""
## {headline}

The package contains a simple end-to-end demo that downloads pre-trained
weights and runs this model on a sample input.

```bash
python -m {model_package}.demo
```

The above demo runs a reference implementation of pre-processing, model
inference, and post processing.

**NOTE**: If you want running in a Jupyter Notebook or Google Colab like
environment, please add the following to your cell (instead of the above).
```
%run -m {model_package}.demo
```
"""


def _get_model_description(
    model_description: str, model_name: str, source_repo: Optional[str]
):
    if source_repo is None:
        return f"{model_description}"
    return f"""
{model_description}

This model is an implementation of {model_name} found [here]({source_repo}).
"""


def _get_demo_ondevice_section(on_device_demo, model_package):
    return (
        f"""
## Run demo on a cloud-hosted device

You can also run the demo on-device.

```bash
python -m {model_package}.demo --on-device
```

**NOTE**: If you want running in a Jupyter Notebook or Google Colab like
environment, please add the following to your cell (instead of the above).
```
%run -m {model_package}.demo -- --on-device
```
"""
        if on_device_demo
        else ""
    )


def _get_how_it_works_section(
    is_precompiled_model,
    model_name,
    repo_url,
    model_package,
    sanitized_components,
    components,
):

    if not components:
        upload_or_compilation_section = (
            _get_compiled_model_upload_section(model_package, sanitized_components)
            if is_precompiled_model
            else _get_compilation_section(model_package)
        )
    else:
        upload_or_compilation_section = _get_compilation_section_with_components(
            model_package, components
        )
    compilation_section = upload_or_compilation_section
    profile_model_intro = _get_profile_model_intro(is_precompiled_model)

    profile_models = _get_profile_models_section(
        is_precompiled_model, sanitized_components, components
    )
    inference_models = _get_inference_models_section(
        is_precompiled_model, sanitized_components, components
    )

    return f"""
## How does this work?

This [export script]({repo_url}/qai_hub_models/models/{model_name}/export.py)
leverages [Qualcomm® AI Hub](https://aihub.qualcomm.com/) to optimize, validate, and deploy this model
on-device. Lets go through each step below in detail:
{compilation_section}


Step 2: **Performance profiling on cloud-hosted device**

{profile_model_intro}. Models can be profiled model on-device using the
`target_model`. Note that this scripts runs the model on a device automatically
provisioned in the cloud.  Once the job is submitted, you can navigate to a
provided job URL to view a variety of on-device performance metrics.
```python
{profile_models}
```

Step 3: **Verify on-device accuracy**

To verify the accuracy of the model on-device, you can run on-device inference
on sample input data on the same cloud hosted device.
```python
{inference_models}
```
With the output of the model, you can compute like PSNR, relative errors or
spot check the output with expected output.

**Note**: This on-device profiling and inference requires access to Qualcomm®
AI Hub. [Sign up for access](https://myaccount.qualcomm.com/signup).
"""


HEADER_FOR_PERF_TABLE = """| Model | Device | Chipset | Target Runtime | Inference Time (ms) | Peak Memory Range (MB) | Precision | Primary Compute Unit | Target Model
|---|---|---|---|---|---|---|---|---|"""

HEADER_FOR_LLM_PERF_TABLE = """| Model | Device | Chipset | Target Runtime | Response Rate (tokens per second) | Time To First Token (range, seconds){eval_row_title}
|---|---|---|---|---|---|"""


def _get_model_purchase_statement(model_info):
    if (
        model_info.llm_details is not None
        and model_info.llm_details.call_to_action
        == LLM_CALL_TO_ACTION.CONTACT_FOR_PURCHASE
    ):
        return "Please contact us to purchase this model."
    return ""


def get_performance_table(
    perf_info: QAIHMModelPerf, is_precompiled=False, has_assets=False
) -> str:
    rows = []
    is_llm = False
    eval_row_name = None
    for model_name, model_perf_detail in perf_info.per_model_details.items():
        # Default to "any" (default) device if it's there. Otherwise, just use the first one the iter gives us.
        for device_perf_defails in model_perf_detail.details_per_device.values():
            device_details = device_perf_defails.device
            for (
                scorecard_path,
                profile_perf_details,
            ) in device_perf_defails.details_per_path.items():
                row = [
                    model_name or perf_info.model_name,
                    device_details.name,
                    device_details.chipset,
                    str(scorecard_path.runtime.name),
                ]

                perf_details = profile_perf_details.perf_details
                if isinstance(perf_details, QAIHMModelPerf.LLMPerformanceDetails):
                    is_llm = True
                    row.extend(
                        [
                            str(perf_details.tokens_per_second),
                            f"{perf_details.time_to_first_token_range_secs[0]} - {perf_details.time_to_first_token_range_secs[1]}",
                        ]
                    )

                    eval_details = profile_perf_details.eval_details
                    if eval_details:
                        if not eval_row_name:
                            eval_row_name = eval_details.name
                        elif eval_row_name != eval_details.name:
                            raise ValueError(
                                "Evaluation metrics for a single model must all have the same name."
                            )
                        row.append(f"{eval_details.value}{eval_details.unit}")
                    else:
                        row.append("--")
                else:
                    mem_min = bytes_to_mb(perf_details.peak_memory_bytes[0])
                    mem_max = bytes_to_mb(perf_details.peak_memory_bytes[1])
                    row.extend(
                        [
                            f"{perf_details.inference_time_microsecs / 1000} ms",
                            f"{mem_min} - {mem_max} MB",
                            perf_details.precision.upper(),
                            perf_details.primary_compute_unit,
                        ]
                    )

                if has_assets:
                    if (
                        scorecard_path.runtime == TargetRuntime.QNN
                        and "8 Gen 2" not in device_details.chipset
                        and "8 Gen 3" not in device_details.chipset
                    ):
                        # We don't upload context binaries for each device.
                        row.append("Use Export Script")
                    else:
                        base_url = ASSET_CONFIG.get_hugging_face_url(
                            perf_info.model_name
                        )
                        ext = None
                        if scorecard_path.runtime == TargetRuntime.TFLITE:
                            ext = "tflite"
                        elif scorecard_path.runtime == TargetRuntime.ONNX:
                            ext = "onnx"
                        elif scorecard_path.runtime == TargetRuntime.QNN:
                            ext = "bin" if is_precompiled else "so"
                        else:
                            raise NotImplementedError()
                        row.append(
                            f"[{perf_info.model_name}.{ext}]({base_url}/blob/main/{model_name}.{ext})"
                        )
                else:
                    row.append("--")

                rows.append("| " + " | ".join(row) + " |")

    if is_llm:
        if eval_row_name:
            header = HEADER_FOR_LLM_PERF_TABLE.replace(
                "{eval_row_title}", f" | {eval_row_name} |"
            )
            header = header + ("---|")
        else:
            header = HEADER_FOR_LLM_PERF_TABLE.replace("{eval_row_title}", "")
    else:
        header = HEADER_FOR_PERF_TABLE

    rows.insert(0, header)
    return "\n".join(rows)


def generate_hf_model_card(
    model_info: QAIHMModelInfo,
    model_perf: QAIHMModelPerf,
    model_code_gen: dict,
    needs_install_instructions: bool,
    model_card: str = MODEL_CARD_TEMPLATE,
    add_export_out: bool = True,
) -> str:
    """Generate a model_card for this model from the template found at hf_model_card_template.md"""

    is_precompiled_model = model_code_gen.get("is_precompiled", False)
    is_aimet_model = model_code_gen.get("is_aimet", False)
    on_device_demo = model_code_gen.get("has_on_target_demo", False)
    has_assets = not model_info.restrict_model_sharing
    model_package = model_info.get_package_name()

    # Replace placeholders
    placeholder_to_replacement_map = [
        [
            "{hugging_face_metadata}",
            str(ruamel.yaml.YAML().dump(model_info.get_hugging_face_metadata())),
        ],
        ["{model_details}", model_info.get_model_details()],
        [
            "{static_image}",
            ASSET_CONFIG.get_web_asset_url(model_info.id, QAIHM_WEB_ASSET.STATIC_IMG),
        ],
        ["{model_definition_url}", str(model_info.get_model_definition_path())],
        ["{demo_url}", str(model_info.get_demo_path())],
        [
            "{package_install_instructions}",
            _get_package_instructions(
                model_info.id, model_info.code_gen_config.pip_install_flags
            )
            if needs_install_instructions
            else PACKAGE_INSTALL_INSTRUCTIONS,
        ],
        ["{configure_instructions}", CONFIGURE_INSTRUCTIONS],
        ["{model_id}", model_info.id],
        ["{model_package}", model_package],
        ["{model_name}", model_info.name],
        [
            "{model_url}",
            f"{AIHUB_URL}/models/{model_info.id}",
        ],
        ["{model_headline}", model_info.headline.strip(".")],
        [
            "{model_description}",
            _get_model_description(
                model_info.description, model_info.name, model_info.source_repo
            ),
        ],
        [
            "{license}",
            _get_licenses(
                model_info.name, model_info.license, model_info.deploy_license
            ),
        ],
        [
            "{reference}",
            _get_references(
                model_info.research_paper_title,
                model_info.research_paper,
                model_info.source_repo,
            ),
        ],
        [
            "{body_perf_table}",
            get_performance_table(model_perf, is_precompiled_model, has_assets),
        ],
        ["{qaism_repo_url}", ASSET_CONFIG.repo_url],
        [
            "{deployment_instructions}",
            _get_deployment_instructions(is_precompiled_model),
        ],
        [
            "{demo_section}",
            _get_demo_section(is_precompiled_model, model_package),
        ],
        [
            "{demo_ondevice_section}",
            _get_demo_ondevice_section(on_device_demo, model_package),
        ],
        [
            "{usage_and_limitations}",
            _get_usage_and_limitations(is_precompiled_model),
        ],
        [
            "{additional_readme_section}",
            model_code_gen.get("additional_readme_section", ""),
        ],
        ["{genie_url}", ASSET_CONFIG.genie_url],
        ["{model_purchase_statement}", _get_model_purchase_statement(model_info)],
    ]

    # Default component is list of components being used
    # components is dictionary of model name -> model object
    default_components = model_code_gen.get("default_components", [])
    components = model_code_gen.get("components", {})
    sanitized_components: dict = {}
    for name, component_obj in components.items():
        if name not in default_components:
            continue
        sanitized_components[name.lower()] = component_obj

    how_it_works = ""
    if not is_aimet_model:
        # TODO: #11126 HF template to update detailed how it works for AIMET model
        how_it_works = _get_how_it_works_section(
            is_precompiled_model,
            model_info.name,
            f"{AIHUB_URL}/models/{model_info.id}",
            model_package,
            sanitized_components,
            components,
        )

    placeholder_to_replacement_map += [
        ["{how_it_works_section}", how_it_works],
        # Skip `Load model` section for pre-compiled models
        # since there is not much to do my loading this model class
        # in python.
        [
            "{load_model_section}",
            _get_load_model_section(model_package) if not is_precompiled_model else "",
        ],
    ]

    # Add placeholders for export specifically
    perf_summary = _get_perf_summary(model_perf) if add_export_out else ""
    placeholder_to_replacement_map += [
        ["{perf_summary}", perf_summary],
    ]

    for placeholder, replacement in placeholder_to_replacement_map:
        model_card = model_card.replace(placeholder, replacement)
    return model_card


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--huggingface-token",
        type=str,
        required=True,
        help="HuggingFace token (write access needed)",
    )
    parser.add_argument(
        "--models",
        type=str,
        help="Comma-separated list of models to update on hugging face. "
        "Each entry should map to one of the folders in qai_hub_models/folders. "
        "All models within the folder that subclass from QAI Hub Models, "
        "will have their assets generated."
        "For example `--models mobilenet_v2,trocr`",
    )
    parser.add_argument(
        "--all",
        action="store_true",
        help="Run on all models.",
    )
    parser.add_argument(
        "--skip-upload-assets",
        action="store_true",
        help="Skip uploading of assets to HuggingFace.",
    )
    parser.add_argument(
        "--public",
        action="store_true",
        help="Make the models specified public on hugging face.",
    )
    return parser.parse_args()


if __name__ == "__main__":
    main()
