# ---------------------------------------------------------------------
# Copyright (c) 2025 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------
import argparse
import os
import re

import adbutils

GENIE_CONFIG_FILE_REL_PATH = "./genie-config.json"


"""
Generating LLM metrics for AI Hub Website

1. Generate model bundles for model of interest
    a. Option A: Use LLM on Genie tutorial https://github.com/quic/ai-hub-apps/tree/main/tutorials/llm_on_genie
    b. Option B: You already have QNN context binary, tokenizer and QNN SDK in place
2. Use this script to run model on-device using genie-t2t-run and collect data
    a. Example usage:
        python generate_llm_metrics.py --bundle-path <bundle_path_from_point_1>  --prompt '<|begin_of_text|><|start_header_id|>system<|end_header_id|> You are a helpful assistant. Answer the following question in consice response.<|eot_id|><|start_header_id|>user<|end_header_id|> What is gravity?<|eot_id|><|start_header_id|>assistant<|end_header_id|> '
    b. You can provide `--adb_port <port>` if want to collect data on QDC device with different port than default i.e. 5037
    c. Provide fix prompt to run this multiple iterations
    d. Set `--keep-on-device-bundle` to keep data on-device. This is useful if you plan to use genie-bundle on device for further use.
3. This outputs profiling data in `perf.yaml` format without device info
    e.g.
        ---------------- LLM Metrics ----------------
        Device details: SoC: SM8750 OS: 15

            llm_metrics:
                time_to_first_token_range:
                min: 85931.5
                max: 2749808.0
                tokens_per_second: 25.77723

    You can copy paste this as is in new section created in perf.yaml for given device.
"""


def _get_processed_kpis(genie_output_log: str) -> tuple[int, float]:
    """
    Get Prompt Processor(PP) and Token Generator(TG) metrics from output log
    """
    prompt_processor_total_time_pattern = r"Prompt Processing Time: (\d+) us"
    token_generator_tps_pattern = r"Token Generation Rate: (\d+.?\d+) toks/sec"

    matches = re.findall(prompt_processor_total_time_pattern, genie_output_log)

    # There could be multiple matches if model output also includes similar pattern
    prompt_processor_time = int(matches[-1])  # total PP time in us

    matches = re.findall(token_generator_tps_pattern, genie_output_log)
    # There could be multiple matches if model output also includes similar pattern
    token_generator_tps = float(
        matches[-1]
    )  # TPS of Token Generator: unit Tokens per second

    return prompt_processor_time, token_generator_tps


def _get_time_to_first_token_range(
    prompt_processor_time: list[int], input_seq_length: int, max_context_length: int
):
    """
    Compute Time To First Token(TTFT):
        If Prompt size is < input seq length of Prompt size.
    """
    # NOTE: if # of tokens in prompt are more than input_seq_length of Prompt Processor (PP),
    # PP runs multiple times to cover full input prompt.
    # <# of times PP runs> = (<# of tokens in input prompt> / <input seq length of PP ) + 1
    #
    # TODO: This can be extended to support prompts > input seuence length with the following:
    # min_prompt_processor_time = sum(prompt_processor_time) / len(prompt_processor_time) / <# of times PP runs>
    # This requires one to use tokenizer for model being used to corretly identify # of tokens
    # Hence, we are not supporting this as of today. Please use shorter prompt for profiling.
    min_prompt_processor_time = sum(prompt_processor_time) / len(prompt_processor_time)
    max_prompt_processor_time = (
        max_context_length / input_seq_length
    ) * min_prompt_processor_time

    return (min_prompt_processor_time, max_prompt_processor_time)


def _get_avg_tokens_per_second(token_generator_tps: list[float]):
    return sum(token_generator_tps) / len(token_generator_tps)


def _get_llm_metrics(
    prompt_processor_time: list[int],
    token_generator_tps: list[float],
    input_seq_length: int,
    max_context_length: int,
):
    """
    Collects and aggregates LLM metrics

        For Prompt Processor, average out time take to response first token and find range for minimum input seq length to max sequence length
        For Token Generator, average out Token per seconds from Genie KPIs
    """

    # Aggregate Time to First Token and TPS
    time_to_first_token_range = _get_time_to_first_token_range(
        prompt_processor_time, input_seq_length, max_context_length
    )
    token_generator_tps = _get_avg_tokens_per_second(token_generator_tps)

    # Write TTFT and TPS in AI Hub Perf yaml format
    llm_metric = f"""
      llm_metrics:
        time_to_first_token_range:
          min: {time_to_first_token_range[0]}
          max: {time_to_first_token_range[1]}
        tokens_per_second: {token_generator_tps:.5f}
"""

    return llm_metric


def _push_bundle_to_device(
    device: adbutils._device.AdbDevice, bundle_path: str, destination_bundle_path: str
):
    for _, dir_names, file_names in os.walk(bundle_path):
        if len(dir_names) > 0:
            raise RuntimeError(
                "Invalid source bundle directory. Expecting no sub-directory in bundle."
            )

        for file_name in file_names:
            full_file_path = str(os.path.join(bundle_path, file_name))
            print(f"Copying {file_name} to device.")
            device.sync.push(full_file_path, destination_bundle_path)


def _run_model_on_device(
    device: adbutils._device.AdbDevice,
    working_dir: str,
    prompt: str,
    num_iterations: int = 5,
    verbose: bool = False,
):
    """
    Runs models on device from working directory
    """
    device_commands = [
        f"cd {working_dir}",
        f"export LD_LIBRARY_PATH={working_dir}",
        f"export ADSP_LIBRARY_PATH={working_dir}",
        f"./genie-t2t-run -c {GENIE_CONFIG_FILE_REL_PATH} -p '{prompt}'",
    ]
    command = ";".join(device_commands)

    prompt_processor_time, token_generator_tps = [], []
    for i in range(num_iterations):
        print(f"\nInference iteration {i + 1}")
        model_output = device.shell(command)

        if verbose:
            print(model_output)

        # Collect KPIs from model output
        pp_time, tg_tps = _get_processed_kpis(str(model_output))
        prompt_processor_time.append(pp_time)
        token_generator_tps.append(tg_tps)

    return prompt_processor_time, token_generator_tps


def main():
    parser = argparse.ArgumentParser(
        "Generate LLM metrics by running model on Android devices using Genie."
    )
    parser.add_argument(
        "--bundle-path",
        type=str,
        default="",
        help="Local path to genie-bundle generated as per https://github.com/quic/ai-hub-apps/tree/main/tutorials/llm_on_genie",
    )
    parser.add_argument(
        "--device-bundle-path",
        type=str,
        default="",
        help="If provided, uses bundle path from device and skips copying data from --bundle-path",
    )
    parser.add_argument(
        "--keep-on-device-bundle",
        action="store_true",
        help="If set, keeps copied on-device bundled as is. Otherwise, cleans up directory after profiling.",
    )
    parser.add_argument(
        "--prompt-processor-input-seq-length",
        type=int,
        default=128,
        help="Input sequence length of Prompt Processor. Check input sequence length of your model by opening Compile or Profile job of Prompt processor's 1st split on AI Hub.",
    )
    parser.add_argument(
        "--max-context-length",
        type=int,
        default=4096,
        help="Provide context length specified during model export. For default values, check AI Hub model details section for given model.",
    )
    parser.add_argument(
        "--num-iterations",
        type=int,
        default=10,
        help="Number of times to run Genie inference for model profiling.",
    )
    parser.add_argument(
        "--adb-port",
        type=int,
        default=5037,
        help="Port to connect adb devices on. Usually 5037, but if using QDC devices, match this with device port you set ssh-tunneling to.",
    )
    parser.add_argument(
        "--prompt",
        type=str,
        required=True,
        help="Refer to prompt format for given model. Try this sample prompt first to ensure output is concise to make profiling efficient, once can achieve this by adding instructions to be concise in the sytem prompt. \n e.g. Refer to the following prompt for Llama3 models: '<|begin_of_text|><|start_header_id|>system<|end_header_id|>\n\nYou are a helpful assistant. Answer the following question in consice response.<|eot_id|><|start_header_id|>user<|end_header_id|>\n\nWhat is gravity?<|eot_id|><|start_header_id|>assistant<|end_header_id|>'",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="If set, dumps genie model inference output for each inference. Otherwise, minimal output of stdout and final metrics.",
    )

    args = parser.parse_args()

    bundle_path = args.bundle_path
    device_bundle_path = args.device_bundle_path
    adb_port = args.adb_port
    input_seq_length = args.prompt_processor_input_seq_length
    max_context_length = args.max_context_length
    num_iterations = args.num_iterations
    prompt = args.prompt
    keep_on_device_bundle = args.keep_on_device_bundle
    verbose = args.verbose

    if bundle_path and device_bundle_path:
        raise RuntimeError(
            "Only one of --bundle-path and --device-bundle-path should be set."
        )

    # Connect to device using ADB client
    adb = adbutils.AdbClient(port=adb_port)
    device = adb.device()

    if bundle_path:
        if not os.path.exists(bundle_path):
            raise RuntimeError(f"Provided --bundle-path {bundle_path} does not exists.")
    elif device_bundle_path:
        if device.shell(f"[ -e {device_bundle_path} ]; echo $?") != "0":
            raise RuntimeError(
                f"Provided --device-bundle-path {device_bundle_path} does not exists on device."
            )
    else:
        raise RuntimeError("One of --bundle-path and --device-bundle-path must be set.")

    device_os = device.prop.get("ro.system.build.version.release")
    device_soc = device.prop.get("ro.soc.model")

    try:
        if device_bundle_path:
            print("\nUsing bundle from device for profiling.")
            working_dir = device_bundle_path
        else:
            working_dir = device.shell("mktemp -d")
            _push_bundle_to_device(device, bundle_path, working_dir)
            print(f"\nBundled directory copied onto device at {working_dir}")

        print(
            f"\nCollecting performance numbers using Genie CLI on {device_soc} with Android OS {device_os}..."
        )

        # Run model on-device using Genie
        prompt_processor_time, token_generator_tps = _run_model_on_device(
            device, working_dir, prompt, num_iterations=num_iterations, verbose=verbose
        )

        # Aggregate LLM metrics
        llm_metric = _get_llm_metrics(
            prompt_processor_time,
            token_generator_tps,
            input_seq_length,
            max_context_length,
        )

        print("\n---------------- LLM Metrics ----------------")
        print(f"Device details: SoC: {device_soc} OS: {device_os}")
        print(llm_metric)

    finally:
        if keep_on_device_bundle or device_bundle_path:
            print(f"Bundle directory from device {working_dir} not removed.")
        else:
            print(f"Removing bundle directory from device {working_dir}.")
            device.shell(f"rm -rf {working_dir}")


if __name__ == "__main__":
    main()
