# ---------------------------------------------------------------------
# Copyright (c) 2024 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------
import os

from qai_hub_models.configs.info_yaml import QAIHMModelInfo
from qai_hub_models.utils.asset_loaders import ASSET_CONFIG, load_yaml
from qai_hub_models.utils.path_helpers import MODEL_IDS
from qai_hub_models.utils.readme_util import (
    _get_licenses,
    _get_package_instructions,
    _get_references,
    _get_usage_and_limitations,
)

README_TEMPLATE = open(
    os.path.join(os.path.dirname(__file__), "templates", "model_readme_template.md")
).read()


def main():
    """Generate a README for each model, and save it in <model_dir>/README.md."""
    for model_id in MODEL_IDS:
        model_info = QAIHMModelInfo.from_model(model_id)
        readme = generate_model_readme(model_info, model_info.has_model_requirements())
        readme_path = model_info.get_readme_path()
        if os.path.exists(readme_path):
            os.remove(readme_path)
        with open(readme_path, "w") as readme_file:
            readme_file.write(readme)


def _get_example_and_usage_section(
    package_install_instructions: str, model_package: str
):
    return f"""
## Example & Usage
{package_install_instructions}

Once installed, run the following simple CLI demo:

```bash
python -m {model_package}.demo
```
More details on the CLI tool can be found with the `--help` option. See
[demo.py](demo.py) for sample usage of the model including pre/post processing
scripts. Please refer to our [general instructions on using
models](../../../#getting-started) for more usage instructions.

## Export for on-device deployment

This repository contains export scripts that produce a model optimized for
on-device deployment. This can be run as follows:

```bash
python -m {model_package}.export
```
Additional options are documented with the `--help` option. Note that the above
script requires access to Deployment instructions for Qualcomm® AI Hub."""


def generate_model_readme(
    model_info: QAIHMModelInfo, needs_install_instructions: bool
) -> str:
    """Generate a README for this model from the template found at .model_readme_template.md"""
    # Replace placeholders
    readme = README_TEMPLATE

    model_code_gen = {}
    model_code_gen_path = model_info.get_code_gen_yaml_path()

    if os.path.exists(model_code_gen_path):
        model_code_gen = load_yaml(model_code_gen_path)

    include_example_and_usage = not model_code_gen.get("skip_example_usage", False)
    additional_info = model_code_gen.get("additional_readme_section", "")
    include_gen_ai_terms = model_info.is_gen_ai_model
    package_installation = (
        _get_package_instructions(
            model_info.id, model_info.code_gen_config.pip_install_flags
        )
        if needs_install_instructions
        else ""
    )
    example_and_usage = _get_example_and_usage_section(
        package_installation, model_info.get_package_name()
    )

    source_repo_details = ""
    if model_info.source_repo:
        source_repo_details = f"This is based on the implementation of {model_info.name} found [here]({model_info.source_repo}). "

    for placeholder, replacement in [
        ["{model_id}", model_info.id],
        ["{additional_info}", additional_info],
        ["{example_and_usage}", example_and_usage if include_example_and_usage else ""],
        ["{model_name}", model_info.name],
        ["{source_repo_details}", source_repo_details],
        [
            "{model_url}",
            f"{ASSET_CONFIG.models_website_url}/models/{model_info.id}"
            if os.path.exists(model_info.get_perf_yaml_path())
            else "#",
        ],
        ["{model_headline}", model_info.headline.strip(".")],
        ["{model_description}", model_info.description],
        [
            "{license}",
            _get_licenses(
                model_info.name, model_info.license, model_info.deploy_license
            ),
        ],
        [
            "{reference}",
            _get_references(
                model_info.research_paper_title,
                model_info.research_paper,
                model_info.source_repo,
            ),
        ],
        [
            "{usage_and_limitations}",
            _get_usage_and_limitations(include_gen_ai_terms),
        ],
        # Replace in additional readme section for LLMs.
        ["{genie_url}", ASSET_CONFIG.genie_url],
    ]:
        readme = readme.replace(placeholder, replacement)
    return readme


if __name__ == "__main__":
    main()
