# ---------------------------------------------------------------------
# Copyright (c) 2024 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------
from __future__ import annotations

import argparse
import copy
import datetime
import multiprocessing
import os
import subprocess
from itertools import cycle

import ruamel.yaml

from qai_hub_models.configs.info_yaml import QAIHMModelInfo
from qai_hub_models.scorecard import ScorecardProfilePath
from qai_hub_models.scorecard.results import PerformanceDiff
from qai_hub_models.scorecard.results.code_gen import update_code_gen_failure_reasons
from qai_hub_models.scorecard.results.spreadsheet import ResultsSpreadsheet
from qai_hub_models.scorecard.results.yaml import (
    COMPILE_YAML_BASE,
    PROFILE_YAML_BASE,
    CompileScorecardJobYaml,
    ProfileScorecardJobYaml,
)
from qai_hub_models.scripts.generate_export_files import generate_export_file
from qai_hub_models.utils.path_helpers import (
    MODEL_IDS,
    QAIHM_MODELS_ROOT,
    SCORECARD_RESULTS_DIR,
)
from qai_hub_models.utils.testing import (
    get_compile_job_ids_file,
    get_profile_job_ids_file,
    is_hub_testing_async,
)


def read_jobs_config(config_path: str) -> dict:
    """Read yaml files."""
    yaml = ruamel.yaml.YAML()
    with open(config_path) as file:
        return yaml.load(file)


def write_jobs_config(config: dict, path: str) -> None:
    """Write yaml files with special characters like copyright logo, etc."""
    yaml = ruamel.yaml.YAML()
    with open(path, "w") as file:
        yaml.dump(config, file)


def remove_failed_jobs(config: dict) -> None:
    """
    Failed jobs need to be in the config to get their job ids for summary but
    we want to delete them before writing to perf.yaml
    """
    for model_config in config["models"]:
        for perf_metrics in model_config["performance_metrics"]:
            for key in list(perf_metrics.keys()):
                if (
                    "job_id" in perf_metrics[key]
                    and perf_metrics[key]["inference_time"] == "null"
                ):
                    del perf_metrics[key]


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--models",
        type=str,
        help="Comma-separated list of models to update. Each entry should map to one of the folders in qai_hub_models/folders.",
        default=os.environ.get("QAIHM_TEST_MODELS", "ALL"),
    )
    parser.add_argument(
        "--profile-ids",
        type=str,
        help="Comma-separated list of paths to additional profile job yamls. Jobs in these YAMLS will be added to or override jobs in the existing profile job intermediates YAML. YAML specified later in the list will override identical jobs in previous YAML.",
    )
    parser.add_argument(
        "--compile-ids",
        type=str,
        help="Comma-separated list of paths to additional compile job yamls. Jobs in these YAMLS will be added to or override jobs in the existing compile job intermediates YAML. YAML specified later in the list will override identical jobs in previous YAML.",
    )
    parser.add_argument(
        "--ignore-existing-intermediate-ids",
        action="store_true",
        help="If set, all existing job IDs under qai_hub_models/scorecard/intermediates/*yaml are ignored. The entire intermediate YAML files will be overwritten with the job IDs provided by --profile-ids and --compile-ids.",
        default=os.environ.get("IGNORE_EXISTING_INTERMEDIATE_JOB_YAMLS", "0") == "1",
    )
    parser.add_argument(
        "--gen-csv",
        action="store_true",
        help="Generate a csv that summarizes the profile and compile steps.",
    )
    parser.add_argument(
        "--gen-perf-summary",
        action="store_true",
        help="Generate a summary of the performance per model, and update perf yaml files.",
    )
    parser.add_argument(
        "--sync-code-gen",
        action="store_true",
        help="Sync code generation YAML with failures & successes in the scorecard YAML. If compile job fails for any device, that path is skipped. If the profile job for the default export device fails, that path is skipped.",
    )
    parser.add_argument("--results-dir", type=str, default=SCORECARD_RESULTS_DIR)
    return parser.parse_args()


def process_model(
    model_id: str,
    compile_job_yamls: CompileScorecardJobYaml,
    profile_job_yamls: ProfileScorecardJobYaml,
    gen_csv: bool,
    sync_code_gen: bool,
    gen_perf_summary: bool,
) -> tuple[ResultsSpreadsheet, list[str], dict | None, dict | None]:
    """
    Returns:
        * Model spreadsheet, or empty spreadsheet if not gen_csv
        * python files modified if code-gen was changed
        * previous (on disk) perf yaml, or None if not gen_perf_summary
        * current (from this scorecard) perf yaml, or None if not gen_perf_summary
    """

    def print_with_id(str):
        print(f"{model_id} | {str}")

    # Load configs
    model_info = QAIHMModelInfo.from_model(model_id)
    cj = model_info.code_gen_config

    # Skip certain models
    if cj.is_precompiled or cj.skip_hub_tests_and_scorecard:
        return ResultsSpreadsheet(), [], None, None

    # Get summaries for this model and its components.
    print_with_id("Loading compile summary")
    compile_summary = compile_job_yamls.summary_from_model(
        model_id, model_info.is_quantized, cj.component_names
    )
    print_with_id("Loading profile summary")
    profile_summary = profile_job_yamls.summary_from_model(
        model_id, model_info.is_quantized, cj.component_names
    )

    entries: ResultsSpreadsheet = ResultsSpreadsheet()
    if gen_csv:
        print_with_id("Adding to Spreadsheet")
        entries.append_model_summary_entries(
            model_info.is_quantized,
            compile_summary,
            profile_summary,
        )

    modified_files = []
    if sync_code_gen:
        # Enable or disable runtimes on this model depending on whether the default device has passing jobs
        if update_code_gen_failure_reasons(
            model_info.is_quantized,
            compile_summary,
            profile_summary,
            cj,
        ):
            code_gen_path = cj.to_model_yaml(model_id)
            modified_files.extend(generate_export_file(model_id))
            print_with_id(f"Updated Runtime Failure Reasons in {code_gen_path}")
        else:
            print_with_id(
                "Runtime Failure Reasons Not Updated (Model did not run in this scorecard)"
            )

    model_card: dict | None = None
    prev_model_card: dict | None = None
    if gen_perf_summary:
        print_with_id("Writing Performance YAML")

        # Determine which scorecard paths are valid (successful) for this model.
        valid_paths = []
        for path in [
            path for path in ScorecardProfilePath if path.include_in_perf_yaml
        ]:
            if cj.supports_runtime(path.runtime):
                valid_paths.append(path)

        # Build model card
        model_card = profile_summary.get_perf_card(
            include_failed_jobs=True,
            include_internal_devices=False,
            exclude_paths=[x for x in ScorecardProfilePath if x not in valid_paths],
            exclude_form_factors=model_info.private_perf_form_factors,
            model_name=model_info.name,
        )

        # Get previous model card
        perf_filename = os.path.join(QAIHM_MODELS_ROOT, model_id, "perf.yaml")
        if os.path.exists(os.path.join(perf_filename)):
            prev_model_card = read_jobs_config(perf_filename)

        # Dump model card without failed jobs
        model_card_copy = copy.deepcopy(model_card)
        remove_failed_jobs(model_card_copy)
        write_jobs_config(model_card_copy, perf_filename)
        print_with_id(f"Wrote {perf_filename}")

    return entries, modified_files, prev_model_card, model_card


if __name__ == "__main__":
    args = parse_args()
    if is_hub_testing_async():
        args.compile_ids = args.compile_ids or str(get_compile_job_ids_file())
        args.profile_ids = args.profile_ids or str(get_profile_job_ids_file())
    os.makedirs(args.results_dir, exist_ok=True)
    now_str = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")

    # List of models for which to generate perf.
    if args.models and args.models.upper() != "ALL":
        model_list = args.models.split(",")
        for model in model_list:
            if model not in MODEL_IDS:
                raise ValueError(f"Unknown model selected: {model}")
    else:
        model_list = MODEL_IDS

    # Load Compile Jobs
    compile_job_yamls = (
        CompileScorecardJobYaml.from_file(COMPILE_YAML_BASE)
        if not args.ignore_existing_intermediate_ids
        else CompileScorecardJobYaml()
    )
    for compile_yaml_path in args.compile_ids.split(",") if args.compile_ids else []:
        compile_job_yamls.update(CompileScorecardJobYaml.from_file(compile_yaml_path))

    # Load Profile Jobs
    profile_job_yamls = (
        ProfileScorecardJobYaml.from_file(PROFILE_YAML_BASE)
        if not args.ignore_existing_intermediate_ids
        else ProfileScorecardJobYaml()
    )
    for profile_yaml_path in args.profile_ids.split(",") if args.profile_ids else []:
        profile_job_yamls.update(ProfileScorecardJobYaml.from_file(profile_yaml_path))

    # Extract Data from Models
    if len(model_list) > 1:
        # Use multiprocessing for multiple models because getting jobs from Hub is slow
        pool = multiprocessing.Pool(processes=15)
        model_summaries = pool.starmap(
            process_model,
            zip(
                model_list,
                cycle([compile_job_yamls]),
                cycle([profile_job_yamls]),
                cycle([args.gen_csv]),
                cycle([args.sync_code_gen]),
                cycle([args.gen_perf_summary]),
            ),
        )
        pool.close()
    else:
        # Single model option for that allows breakpoints
        model_summaries = [
            process_model(
                model_list[0],
                compile_job_yamls,
                profile_job_yamls,
                args.gen_csv,
                args.sync_code_gen,
                args.gen_perf_summary,
            )
        ]

    # Collect Model Results
    perf_report = PerformanceDiff() if args.gen_perf_summary else None
    spreadsheet = ResultsSpreadsheet(has_inference_jobs=False) if args.gen_csv else None
    modified_files = []
    for model_id, (
        model_spreadsheet,
        model_code_gen_files,
        prev_model_card,
        curr_model_card,
    ) in zip(model_list, model_summaries):
        # Combine model spreadsheet with group spreadsheet
        if spreadsheet is not None:
            spreadsheet.extend(model_spreadsheet)

        # Update code generated files
        modified_files.extend(model_code_gen_files)

        # Update performance report with model card diff
        if perf_report is not None:
            if curr_model_card is not None:
                # Summary is made between the existing perf.yaml and the newly
                # created model card.
                perf_report.update_summary(
                    model_id,
                    previous_report=prev_model_card,
                    new_report=curr_model_card,
                )
            elif prev_model_card is not None:
                perf_report.add_missing_model(model_id)

    # Write spreadsheet to disk
    if spreadsheet is not None:
        summary_path = os.path.join(
            args.results_dir, f"scorecard-summary-{now_str}.csv"
        )
        spreadsheet.to_csv(summary_path)
        print(f"Spreadsheet written to {os.path.realpath(summary_path)}")

    # Write summary to disk
    if perf_report is not None:
        report_path = os.path.join(
            args.results_dir, f"performance-summary-{now_str}.txt"
        )
        perf_report.dump_summary(report_path)
        compile_job_yamls.to_file(COMPILE_YAML_BASE)
        profile_job_yamls.to_file(PROFILE_YAML_BASE)
        print(f"Compile Job IDs written to {COMPILE_YAML_BASE}")
        print(f"Profile Job IDs written to {PROFILE_YAML_BASE}")

    # Run pre-commit on re-generated files
    if modified_files:
        subprocess.run(["pre-commit", "run", "--files"] + modified_files)
