# ---------------------------------------------------------------------
# Copyright (c) 2024 Qualcomm Innovation Center, Inc. All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# ---------------------------------------------------------------------
import os

import pytest
from qai_hub.client import Client

from qai_hub_models.configs.info_yaml import QAIHMModelInfo
from qai_hub_models.scorecard.results.yaml import ProfileScorecardJobYaml

EXPECTED_MODEL_CARD = {
    "aggregated": {
        "supported_devices": [
            "Samsung Galaxy S24",
            "Samsung Galaxy S24 Ultra",
            "Samsung Galaxy S24+",
            "Samsung Galaxy S23",
            "Samsung Galaxy S23 Ultra",
            "Samsung Galaxy S23+",
            "Samsung Galaxy S22 5G",
            "Samsung Galaxy S22 Ultra 5G",
            "Samsung Galaxy S22+ 5G",
            "Samsung Galaxy Tab S8",
            "Xiaomi 12",
            "Xiaomi 12 Pro",
            "Samsung Galaxy S21",
            "Samsung Galaxy S21 Ultra",
            "Samsung Galaxy S21+",
        ],
        "supported_chipsets": [
            "Snapdragon® 8 Gen 3",
            "Snapdragon® 8 Gen 2",
            "Snapdragon® 8 Gen 1",
            "Snapdragon® 888",
        ],
    },
    "models": [
        {
            "name": "TrOCRDecoder",
            "universal_assets": {
                "torchscript_onnx": "mkqpd0woq",
                "torchscript_onnx_tflite": "m0q9wd4rm",
            },
            "performance_metrics": [
                {
                    "torchscript_onnx_tflite": {
                        "inference_time": 2694.0,
                        "throughput": 371.19524870081665,
                        "estimated_peak_memory_range": {"min": 12288, "max": 2469896},
                        "primary_compute_unit": "NPU",
                        "precision": "fp16",
                        "layer_info": {
                            "layers_on_npu": 370,
                            "layers_on_gpu": 0,
                            "layers_on_cpu": 0,
                            "total_layers": 370,
                        },
                        "job_id": "jegnl0nj5",
                        "job_status": "Passed",
                    },
                    "torchscript_onnx": {
                        "inference_time": 2880.0,
                        "throughput": 347.22222222222223,
                        "estimated_peak_memory_range": {"min": 0, "max": 484232016},
                        "primary_compute_unit": "NPU",
                        "precision": "fp16",
                        "layer_info": {
                            "layers_on_npu": 1,
                            "layers_on_gpu": 0,
                            "layers_on_cpu": 1,
                            "total_layers": 2,
                        },
                        "job_id": "j2p034m0p",
                        "job_status": "Passed",
                    },
                    "reference_device_info": {
                        "name": "Samsung Galaxy S23",
                        "os": "13",
                        "form_factor": "Phone",
                        "os_name": "Android",
                        "manufacturer": "Samsung",
                        "chipset": "Snapdragon® 8 Gen 2",
                    },
                },
                {
                    "torchscript_onnx_tflite": {
                        "inference_time": 1922.0,
                        "throughput": 520.2913631633714,
                        "estimated_peak_memory_range": {"min": 12288, "max": 192790624},
                        "primary_compute_unit": "NPU",
                        "precision": "fp16",
                        "layer_info": {
                            "layers_on_npu": 370,
                            "layers_on_gpu": 0,
                            "layers_on_cpu": 0,
                            "total_layers": 370,
                        },
                        "job_id": "jep20xw6g",
                        "job_status": "Passed",
                    },
                    "torchscript_onnx": {
                        "inference_time": 2363.0,
                        "throughput": 423.1908590774439,
                        "estimated_peak_memory_range": {"min": 0, "max": 36730784},
                        "primary_compute_unit": "NPU",
                        "precision": "fp16",
                        "layer_info": {
                            "layers_on_npu": 1,
                            "layers_on_gpu": 0,
                            "layers_on_cpu": 1,
                            "total_layers": 2,
                        },
                        "job_id": "jogk7v2vp",
                        "job_status": "Passed",
                    },
                    "reference_device_info": {
                        "name": "Samsung Galaxy S24",
                        "os": "14",
                        "form_factor": "Phone",
                        "os_name": "Android",
                        "manufacturer": "Samsung",
                        "chipset": "Snapdragon® 8 Gen 3",
                    },
                },
            ],
        },
        {
            "name": "TrOCREncoder",
            "universal_assets": {
                "torchscript_onnx": "m1q8z1w3q",
                "torchscript_onnx_tflite": "mpn74pd4n",
            },
            "performance_metrics": [
                {
                    "torchscript_onnx_tflite": {
                        "inference_time": 216726.0,
                        "throughput": 4.614121056080028,
                        "estimated_peak_memory_range": {
                            "min": 6426624,
                            "max": 10110760,
                        },
                        "primary_compute_unit": "NPU",
                        "precision": "fp16",
                        "layer_info": {
                            "layers_on_npu": 592,
                            "layers_on_gpu": 0,
                            "layers_on_cpu": 0,
                            "total_layers": 592,
                        },
                        "job_id": "jo5mqem7p",
                        "job_status": "Passed",
                    },
                    "torchscript_onnx": {
                        "inference_time": 188685.0,
                        "throughput": 5.299838354930174,
                        "estimated_peak_memory_range": {
                            "min": 14303232,
                            "max": 141368616,
                        },
                        "primary_compute_unit": "NPU",
                        "precision": "fp16",
                        "layer_info": {
                            "layers_on_npu": 1,
                            "layers_on_gpu": 0,
                            "layers_on_cpu": 0,
                            "total_layers": 1,
                        },
                        "job_id": "jqpyrzx05",
                        "job_status": "Passed",
                    },
                    "reference_device_info": {
                        "name": "Samsung Galaxy S23",
                        "os": "13",
                        "form_factor": "Phone",
                        "os_name": "Android",
                        "manufacturer": "Samsung",
                        "chipset": "Snapdragon® 8 Gen 2",
                    },
                },
                {
                    "torchscript_onnx_tflite": {
                        "inference_time": 163291.0,
                        "throughput": 6.124036229798335,
                        "estimated_peak_memory_range": {"min": 16384, "max": 317166800},
                        "primary_compute_unit": "NPU",
                        "precision": "fp16",
                        "layer_info": {
                            "layers_on_npu": 592,
                            "layers_on_gpu": 0,
                            "layers_on_cpu": 0,
                            "total_layers": 592,
                        },
                        "job_id": "jopr860k5",
                        "job_status": "Passed",
                    },
                    "torchscript_onnx": {
                        "inference_time": 144066.0,
                        "throughput": 6.941263032221343,
                        "estimated_peak_memory_range": {
                            "min": 14761984,
                            "max": 91752272,
                        },
                        "primary_compute_unit": "NPU",
                        "precision": "fp16",
                        "layer_info": {
                            "layers_on_npu": 1,
                            "layers_on_gpu": 0,
                            "layers_on_cpu": 0,
                            "total_layers": 1,
                        },
                        "job_id": "j1p802eqg",
                        "job_status": "Passed",
                    },
                    "reference_device_info": {
                        "name": "Samsung Galaxy S24",
                        "os": "14",
                        "form_factor": "Phone",
                        "os_name": "Android",
                        "manufacturer": "Samsung",
                        "chipset": "Snapdragon® 8 Gen 3",
                    },
                },
            ],
        },
    ],
}


def test_generate_perf():
    # Verify the model card is made correctly given the info
    api_url = Client().config.api_url
    if not api_url.startswith("https://app.aihub.qualcomm.com"):
        pytest.skip(
            f"This test uses jobs only accessible on production AI Hub. Enabled URL: {api_url}"
        )

    job_ids = ProfileScorecardJobYaml.from_file(
        os.path.join(os.path.dirname(__file__), "profile_job_ids.yaml")
    )
    model_info = QAIHMModelInfo.from_model("trocr")
    model_perf = job_ids.summary_from_model(
        model_info.id,
        model_info.is_quantized,
        model_info.code_gen_config.component_names,
    )
    model_card = model_perf.get_perf_card()

    # Erase the timestamp, since it will vary
    for model in model_card["models"]:
        for metric in model["performance_metrics"]:
            del metric["timestamp"]

    assert model_card == EXPECTED_MODEL_CARD
